# type: ignore
import os
import pickle
from pathlib import Path
from typing import Any, Callable, List, Tuple

import keras
import keras_nlp
import numpy as np
import tensorflow as tf
import tensorflow_datasets as tfds
import torch
from keras.saving import deserialize_keras_object, serialize_keras_object
from keras_nlp.samplers import Sampler
from numpy.typing import NDArray
from tensorflow.python.framework.ops import EagerTensor
from torch import nn
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm

from omnivault.transformer.config.composer import Composer
from omnivault.transformer.config.criterion import CRITERION_REGISTRY
from omnivault.transformer.config.decoder import (
    AddNormConfig,
    DecoderBlockConfig,
    DecoderConfig,
    MultiHeadedAttentionConfig,
    PositionwiseFeedForwardConfig,
)
from omnivault.transformer.config.generator import GeneratorConfig
from omnivault.transformer.config.optim import OPTIMIZER_REGISTRY
from omnivault.transformer.config.scheduler import SCHEDULER_REGISTRY
from omnivault.transformer.config.trainer import TrainerConfig
from omnivault.transformer.core.dataset import (
    construct_dummy_batch_future_masks,
    construct_dummy_batch_target_padding_masks,
)
from omnivault.transformer.core.state import State
from omnivault.transformer.core.trainer import Trainer, TrainerEvent
from omnivault.transformer.decoder.core import GPTDecoder
from omnivault.transformer.modules.attention.core import ScaledDotProductAttention
from omnivault.transformer.utils.visualization import save_plot_history
from omnivault.utils.reproducibility.seed import seed_all
from omnivault.utils.torch_utils.cleanup import purge_global_scope

seed_all(set_torch_deterministic=False)  # set to False since it may cause a slight increase in memory usage
tf.random.set_seed(1992)

# Paths
DATA_URL: str = "https://dldata-public.s3.us-east-2.amazonaws.com/simplebooks.zip"
BASE_DATA_DIR: str = os.path.join(os.path.expanduser("~/.keras/datasets/"), "simplebooks/")
TRAIN_FILE: str = os.path.join(BASE_DATA_DIR, "simplebooks-92-raw/train.txt")
VALID_FILE: str = os.path.join(BASE_DATA_DIR, "simplebooks-92-raw/valid.txt")

TRAIN_DATASET_PICKLED_PATH = "/kaggle/input/simplebooks-92/train_dataset.pkl"
VALID_DATASET_PICKLED_PATH = "/kaggle/input/simplebooks-92/valid_dataset.pkl"
SAVE_DATASET_AS_PICKLE = False
LOAD_DATASET_FROM_PICKLE = True
COMPUTE_WORD_PIECE_VOCABULARY = False


# Data
BATCH_SIZE = 64
MIN_STRING_LEN = 512  # Strings shorter than this will be discarded
SEQ_LEN = 128  # Length of training sequences, in tokens
CURRENT_SCRIPT_DIR = Path(__file__).resolve().parent

# Vocabulary
VOCABULARY_SERIALIZABLED: str = CURRENT_SCRIPT_DIR / "./assets/vocabulary_serializabled.pkl"
VOCAB_SIZE = 5000  # Limits parameters in model.

# Training
GRADIENT_ACCUMULATION_STEPS = 4
BASE_LR = 0.00025
BASE_TRAIN_BATCH_SIZE = 16
NUM_GPUS = 1

# Inference
NUM_TOKENS_TO_GENERATE = 80


def download_and_extract_data(url: str) -> str:
    """Download and extract dataset from a given URL."""
    return keras.utils.get_file(origin=url, extract=True)


def load_and_prepare_dataset(
    file_path: str, batch_size: int, min_string_len: int, shuffle: bool = False
) -> tf.data.TextLineDataset:
    """Prepare the dataset by loading, filtering and batching."""
    dataset = tf.data.TextLineDataset(file_path)
    dataset = dataset.filter(lambda x: tf.strings.length(x) > min_string_len)
    dataset = dataset.batch(batch_size)
    if shuffle:
        dataset = dataset.shuffle(buffer_size=256)
    return dataset


def load_vocab(vocab_file: str) -> List[str]:
    """Load the vocabulary from a serialized file."""
    try:
        with open(vocab_file, "rb") as f:
            vocab_config = pickle.load(f)
    except FileNotFoundError:
        raise Exception(f"Vocabulary file not found: {vocab_file}") from None
    except Exception as err:
        raise Exception(f"Error loading vocabulary: {str(err)}") from err

    return deserialize_keras_object(vocab_config)


def create_tokenizer(vocabulary: List[str], seq_len: int) -> keras_nlp.tokenizers.WordPieceTokenizer:
    """
    Create a WordPiece tokenizer and a StartEnd packer using the given
    vocabulary and sequence length.
    """
    tokenizer = keras_nlp.tokenizers.WordPieceTokenizer(
        vocabulary=vocabulary,
        sequence_length=seq_len,
        lowercase=True,
    )

    return tokenizer


def create_start_packer(tokenizer: keras_nlp.tokenizers.WordPieceTokenizer) -> keras_nlp.layers.StartEndPacker:
    """Create a StartEnd packer using the given tokenizer."""
    return keras_nlp.layers.StartEndPacker(
        sequence_length=tokenizer.sequence_length,
        start_value=tokenizer.token_to_id("[BOS]"),
    )


def preprocess(inputs: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor]:
    """
    Preprocess the input data by tokenizing and then packing with start tokens.

    Parameters
    ----------
    inputs : tf.Tensor
        Input text data in tensor format.

    Returns
    -------
    Tuple[tf.Tensor, tf.Tensor]
        A tuple containing:
        - features: Tokenized and packed input sequences.
        - labels: Tokenized output sequences (identical to `outputs` from tokenizer).
    """
    outputs = tokenizer(inputs)
    features = start_packer(outputs)
    labels = outputs
    return features, labels


def prepare_and_preprocess_dataset(
    dataset: tf.data.Dataset, preprocess_fn: Callable[[tf.Tensor], Tuple[tf.Tensor, tf.Tensor]]
) -> tf.data.Dataset:
    """
    Prepares the dataset by applying preprocessing, batching, and prefetching.

    Parameters
    ----------
    dataset: tf.data.Dataset
        The raw dataset to be processed.
    preprocess_fn: callable
        The preprocessing function to be applied to the dataset.
    batch_size: int
        The size of batches to divide the dataset into.

    Returns
    -------
    tf.data.Dataset
        The processed and batched dataset.
    """
    return dataset.map(preprocess_fn, num_parallel_calls=tf.data.AUTOTUNE).prefetch(tf.data.AUTOTUNE)


def flatten_dataset(dataset: List[Tuple[np.ndarray, np.ndarray]]) -> List[Tuple[np.ndarray, np.ndarray]]:
    """
    Flatten each pair of source and target arrays in the dataset from shape [B, L] to [B*L].

    Parameters
    ----------
    dataset : List[Tuple[np.ndarray, np.ndarray]]
        The dataset to be flattened, consisting of tuples of source and target numpy arrays.

    Returns
    -------
    List[Tuple[np.ndarray, np.ndarray]]
        The flattened dataset, where each source and target array is reshaped to [B*L].
    """
    flattened_data = [
        (source[i].reshape(-1), target[i].reshape(-1))
        for source, target in tqdm(dataset)
        for i in range(source.shape[0])
    ]
    return flattened_data


class TFDatasetWrapper(Dataset):
    def __init__(self, tf_dataset_as_numpy: List[int]) -> None:
        super().__init__()
        self.tf_dataset_as_numpy = tf_dataset_as_numpy

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, torch.Tensor]:
        source, target = self.tf_dataset_as_numpy[index]

        # Convert from list to tensors
        source = torch.tensor(source, dtype=torch.long)
        target = torch.tensor(target, dtype=torch.long)
        return source, target

    def __len__(self) -> int:
        return len(self.tf_dataset_as_numpy)


def custom_collate_fn(
    batch: List[Tuple[torch.Tensor, torch.Tensor]]
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    sources, targets = zip(*batch)

    sources = torch.stack(sources)
    targets = torch.stack(targets)

    batch_size, seq_len = targets.size(0), targets.size(1)

    future_masks = construct_dummy_batch_future_masks(batch_size, seq_len)
    target_padding_masks = construct_dummy_batch_target_padding_masks(batch_size, seq_len)

    return sources, targets, future_masks, target_padding_masks


def next(prompt: tf.Tensor, cache: Tuple[Any, ...], index: EagerTensor) -> Tuple[tf.Tensor, None, Tuple[Any, ...]]:
    prompt: NDArray[np.int32] = prompt.numpy()
    prompt: torch.LongTensor = torch.from_numpy(prompt).to(composer.trainer.device)

    index: int = int(index)
    with torch.no_grad():
        logits: torch.Tensor = model(prompt)[:, index - 1, :]
    logits: NDArray[float] = logits.detach().cpu().numpy()
    logits: tf.Tensor = tf.convert_to_tensor(logits)
    # Ignore hidden states for now; only needed for contrastive search.
    hidden_states = None
    return logits, hidden_states, cache


@torch.no_grad()
def generate_on_train_epoch_end(trainer: Trainer) -> None:
    """Evaluate and generate on train batch end."""

    model = trainer.model
    model.eval()

    def get_samplers() -> List[Sampler]:
        samplers = [
            keras_nlp.samplers.GreedySampler(temperature=1.0),
            keras_nlp.samplers.BeamSampler(num_beams=10, temperature=1.0),
            keras_nlp.samplers.RandomSampler(temperature=1.0),
            keras_nlp.samplers.TopKSampler(k=20, temperature=0.7),
            keras_nlp.samplers.TopPSampler(p=0.5, temperature=0.7, k=20),
        ]
        return samplers

    samplers = get_samplers()
    tokenizer = trainer.state.tokenizer

    start_packer = create_start_packer(tokenizer)
    # The "packer" layers adds the [BOS] token for us.
    prompt_tokens = start_packer(tokenizer([""]))

    for sampler in samplers:
        generated_tokens = sampler(
            next=next,
            prompt=prompt_tokens,
            index=1,  # Start sampling immediately after the [BOS] token.
        )
        generated_tokens_decoded = tokenizer.detokenize(generated_tokens)
        sampler_name = sampler.__class__.__name__
        trainer.logger.info("%s search Generated text %s", sampler_name, generated_tokens_decoded)
    # Revert model to training mode
    model.train()


if __name__ == "__main__":
    download_and_extract_data(DATA_URL)
    raw_train_ds = load_and_prepare_dataset(TRAIN_FILE, BATCH_SIZE, MIN_STRING_LEN, shuffle=True)
    raw_valid_ds = load_and_prepare_dataset(VALID_FILE, BATCH_SIZE, MIN_STRING_LEN, shuffle=False)

    if COMPUTE_WORD_PIECE_VOCABULARY:
        vocabulary = keras_nlp.tokenizers.compute_word_piece_vocabulary(
            raw_train_ds,
            vocabulary_size=VOCAB_SIZE,
            lowercase=True,
            reserved_tokens=["[PAD]", "[UNK]", "[BOS]"],
        )
        with open(VOCABULARY_SERIALIZABLED, "wb") as f:
            pickle.dump(serialize_keras_object(vocabulary), f)
    else:
        vocabulary = load_vocab(VOCABULARY_SERIALIZABLED)

    tokenizer = create_tokenizer(vocabulary, SEQ_LEN)
    start_packer = create_start_packer(tokenizer)

    train_ds = prepare_and_preprocess_dataset(raw_train_ds, preprocess)
    valid_ds = prepare_and_preprocess_dataset(raw_valid_ds, preprocess)

    if LOAD_DATASET_FROM_PICKLE:
        with open(TRAIN_DATASET_PICKLED_PATH, "rb") as file:
            train_ds_as_list_of_numpy = pickle.load(file)

        with open(VALID_DATASET_PICKLED_PATH, "rb") as file:
            valid_ds_as_list_of_numpy = pickle.load(file)

    if SAVE_DATASET_AS_PICKLE and not LOAD_DATASET_FROM_PICKLE:
        train_ds_as_list_of_numpy = list(tfds.as_numpy(train_ds))
        valid_ds_as_list_of_numpy = list(tfds.as_numpy(valid_ds))

        with open(TRAIN_DATASET_PICKLED_PATH, "wb") as file:
            pickle.dump(train_ds_as_list_of_numpy, file)

        with open(VALID_DATASET_PICKLED_PATH, "wb") as file:
            pickle.dump(valid_ds_as_list_of_numpy, file)

    train_ds_as_numpy = flatten_dataset(train_ds_as_list_of_numpy)
    valid_ds_as_numpy = flatten_dataset(valid_ds_as_list_of_numpy)

    train_ds_pytorch = TFDatasetWrapper(train_ds_as_numpy)
    valid_ds_pytorch = TFDatasetWrapper(valid_ds_as_numpy)

    # Create a PyTorch data loader
    train_loader = DataLoader(
        train_ds_pytorch, shuffle=False, batch_size=16, collate_fn=custom_collate_fn
    )  # shuffled by tensorflow earlier
    valid_loader = DataLoader(
        valid_ds_pytorch, shuffle=False, batch_size=32, collate_fn=custom_collate_fn
    )  # shuffled by tensorflow earlier

    masked_self_attention_mha_config = MultiHeadedAttentionConfig(
        attention=ScaledDotProductAttention(), d_model=256, H=4, dropout=0.2
    )

    feed_forward_config = PositionwiseFeedForwardConfig(
        d_model=256, d_ff=256 * 4, activation=nn.GELU(approximate="tanh"), dropout=0.2, bias=True
    )

    add_norm_config_1 = AddNormConfig(feature_dim=256, dropout=0.2)
    add_norm_config_2 = AddNormConfig(feature_dim=256, dropout=0.2)

    # Create DecoderBlockConfig
    decoder_block_config = DecoderBlockConfig(
        masked_self_attention_mha=masked_self_attention_mha_config,
        feed_forward=feed_forward_config,
        add_norm_1=add_norm_config_1,
        add_norm_2=add_norm_config_2,
    )

    # Create the overall DecoderConfig
    model_config = DecoderConfig(
        d_model=256,
        vocab_size=len(vocabulary),
        context_length=SEQ_LEN,
        num_decoder_blocks=2,
        dropout=0.2,
        decoder_block=decoder_block_config,
    )

    optimizer_config_cls = OPTIMIZER_REGISTRY["torch.optim.Adam"]
    optimizer_pydantic_config = optimizer_config_cls(name="torch.optim.Adam", lr=BASE_LR * GRADIENT_ACCUMULATION_STEPS)

    criterion_config_cls = CRITERION_REGISTRY["torch.nn.CrossEntropyLoss"]
    criterion_pydantic_config = criterion_config_cls(name="torch.nn.CrossEntropyLoss")

    scheduler_config_cls = SCHEDULER_REGISTRY["torch.optim.lr_scheduler.CosineAnnealingLR"]
    scheduler_pydantic_config = scheduler_config_cls(name="torch.optim.lr_scheduler.CosineAnnealingLR", T_max=8)

    trainer_config = TrainerConfig(
        device="cuda",
        max_epochs=8,
        eval_every_n_steps=10000,
        log_every_n_steps=10000,
        use_amp=True,
        autocast_config={"enabled": True, "dtype": torch.float16, "cache_enabled": True},
        scaler_config={
            "enabled": True,
            "init_scale": 2.0**16,
            "growth_factor": 2.0,
            "backoff_factor": 0.5,
            "growth_interval": 2000,
        },
        gradient_accumulation_steps=GRADIENT_ACCUMULATION_STEPS,
        clip_grad_norm={"max_norm": 1.0, "norm_type": 2.0, "error_if_nonfinite": False, "foreach": None},
        apply_weight_decay_to_different_param_groups=False,
        step_scheduler_on_batch_or_epoch="epoch",
        save_dir="./data/simplybooks92/checkpoints",
        save_every_epoch=False,
        save_best_only=True,
        monitor="valid_this_epoch_average_loss",
        mode="min",
    )

    generator_config = GeneratorConfig(temperature=1.0, max_tokens=80, greedy=False, top_k=10, top_p=None)

    composer = Composer(
        model=model_config,
        optimizer=optimizer_pydantic_config,
        criterion=criterion_pydantic_config,
        scheduler=scheduler_pydantic_config,
        trainer=trainer_config,
        generator=generator_config,
    )
    model = GPTDecoder(model_config)
    model = model.to(device=composer.trainer.device, dtype=next(model.parameters()).dtype, non_blocking=True)

    optimizer = optimizer_pydantic_config.build(params=model.parameters())
    criterion = criterion_pydantic_config.create_instance()

    composer.scheduler = scheduler_pydantic_config
    scheduler = scheduler_pydantic_config.build(optimizer=optimizer)

    composer.pretty_print()

    state = State(
        model=model,
        criterion=criterion,
        optimizer=optimizer,
        scheduler=scheduler,
        vocabulary=vocabulary,
        tokenizer=tokenizer,
    )

    device = composer.trainer.device
    trainer = Trainer(
        state=state,
        composer=composer,
        logger=None,
        device=device,  # type: ignore[arg-type]
    )
    trainer.add_callback(TrainerEvent.ON_TRAIN_EPOCH_END.value, generate_on_train_epoch_end)

    _trained_state = trainer.fit(train_loader=train_loader, valid_loader=valid_loader)
    _trained_state.pretty_print()
    history = _trained_state.history
    _ = save_plot_history(history, plot=False, save_path=f"{composer.trainer.save_dir}/history.png")

    purge_global_scope(["optimizer", "state", "trainer"])



from __future__ import annotations

from pathlib import Path


def find_root_dir(current_path: Path | None = None, marker: str = ".git") -> Path | None:
    """
    Find the root directory by searching for a directory or file that serves as a
    marker.

    Parameters
    ----------
    current_path : Path | None
        The starting path to search from. If None, the current working directory
        `Path.cwd()` is used.
    marker : str
        The name of the file or directory that signifies the root.

    Returns
    -------
    Path | None
        The path to the root directory. Returns None if the marker is not found.
    """
    if not current_path:
        current_path = Path.cwd()
    current_path = current_path.resolve()
    for parent in [current_path, *current_path.parents]:
        if (parent / marker).exists():
            return parent
    return None



from enum import Enum

import torch


class Reduction(Enum):
    NONE = "none"
    MEAN = "mean"
    SUM = "sum"


class CrossEntropyLoss:
    def __init__(self, reduction: Reduction = Reduction.MEAN) -> None:
        """
        Initialize the CrossEntropyLoss.

        Parameters
        ----------
        reduction : Reduction, optional
            Specifies the reduction to apply to the output.
            Defaults to Reduction.MEAN.
        """
        self.reduction = reduction

    def __call__(self, logits: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        """
        Compute the cross entropy loss for a given set of predicted logits and targets.

        Parameters
        ----------
        logits : torch.Tensor
            Predicted logits tensor of shape (batch_size, ..., vocab_size)
            where ... represents an arbitrary number of dimensions. Note
            `vocab_size` is the last dimension but is used in context of transformer models.
        targets : torch.Tensor
            Target tensor of shape (batch_size, ...).

        Returns
        -------
        torch.Tensor
            The computed cross entropy loss, reduced according to the specified reduction strategy.
        """
        if logits.shape[:-1] != targets.shape:
            raise ValueError(
                f"Logits and targets must have compatible shapes. "
                f"Received logits shape: {logits.shape} and targets shape: {targets.shape}"
            )

        if len(logits.shape) > 3:
            raise ValueError(f"Only 1D, 2D, and 3D logits are supported. " f"Received logits shape: {logits.shape}")

        logits_max, _ = torch.max(logits, dim=-1, keepdim=True)
        logits_stable = logits - logits_max

        log_sum_exp = torch.log(torch.sum(torch.exp(logits_stable), dim=-1))
        target_logits = logits_stable.gather(dim=-1, index=targets.unsqueeze(-1)).squeeze(-1)
        neg_log_likelihood = -target_logits + log_sum_exp

        if self.reduction == Reduction.NONE:
            return neg_log_likelihood
        elif self.reduction == Reduction.MEAN:
            return torch.mean(neg_log_likelihood)
        elif self.reduction == Reduction.SUM:
            return torch.sum(neg_log_likelihood)
        else:
            raise ValueError(f"Invalid reduction: {self.reduction}")



import torch

from omnivault.transformer.core.trainer import Trainer


@torch.no_grad()
def evaluate_generate_on_train_batch_end(trainer: Trainer) -> None:
    """Evaluate and generate on train batch end."""

    tokenizer = trainer.state.tokenizer
    generator_config = trainer.composer.generator

    if trainer.train_batch_index % trainer.eval_every_n_steps == 0:
        # Perform evaluation and sample generation
        model = trainer.model
        model.eval()

        context = "O God, O God!"
        starting_tokens = tokenizer.encode(context)
        starting_tokens = torch.tensor(starting_tokens, dtype=torch.long, device=trainer.device)  # type: ignore[assignment]

        generated_tokens = model.generate(starting_tokens, **generator_config.model_dump(mode="python"))
        generated_tokens = generated_tokens.squeeze(0)
        generated_tokens_decoded = tokenizer.decode(encoded_sequence=generated_tokens, remove_special_tokens=True)
        trainer.logger.info("Generated text %s", generated_tokens_decoded)

        # Revert model to training mode
        model.train()



from typing import Iterable

import torch
from torch import nn

__all__ = ["gradient_clipping"]


def gradient_clipping(parameters: Iterable[nn.Parameter], max_norm: float, epsilon: float = 1e-6) -> None:
    """
    Clips the gradients of the provided parameters to have a maximum L2 norm of `max_norm`.

    Parameters
    ----------
    parameters : Iterable[nn.Parameter]
        An iterable of PyTorch parameters.
    max_norm : float
        The maximum L2 norm for gradient clipping.
    epsilon : float, optional
        A small value added for numerical stability. Defaults to 1e-6.
    """
    parameters_with_grad = [p for p in parameters if p.grad is not None]

    # l2 norm of the gradients
    grad_norms = torch.stack([torch.norm(p.grad, 2) for p in parameters_with_grad])
    total_norm = torch.norm(grad_norms, 2)

    if total_norm > max_norm:
        clip_coef = max_norm / (total_norm + epsilon)
        for p in parameters_with_grad:
            assert p.grad is not None  # appease mypy
            p.grad.mul_(clip_coef)



from __future__ import annotations

import logging
from typing import Type

from pydantic import BaseModel, field_validator


class DistributedConfig(BaseModel):
    """Default settings, override if ddp, else stay as is for non-ddp."""

    log_dir: str = "logs_distributed"
    log_level: int = logging.INFO
    log_on_master_or_all: bool = True
    master_addr: str = "localhost"
    master_port: str = "29500"
    nnodes: int = 1
    nproc_per_node: int = 1
    node_rank: int = 0
    world_size: int = 1
    backend: str = "gloo"
    init_method: str = "env://"

    @field_validator("master_port", mode="plain")
    @classmethod
    def coerce_str_to_int(cls: Type[DistributedConfig], v: str) -> int:
        return int(v)



from __future__ import annotations

from typing import Any, Dict, List, Set, Tuple, TypedDict

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import torch
import torch.optim as optim
from torch import nn

__all__ = [
    "total_trainable_parameters",
    "total_parameters",
    "compare_models",
    "compare_models_and_report_differences",
    "get_named_modules",
    "gather_weight_stats",
    "prepare_stats_dataframe",
    "plot_distribution_stats",
    "get_named_parameters",
    "Freezer",
    "freeze_layers",
    "check_optimizer_coverage",
    "get_model_layer_info",
]


def total_trainable_parameters(module: nn.Module) -> int:
    """Returns the number of trainable parameters in the model."""
    return sum(p.numel() for p in module.parameters() if p.requires_grad)


def total_parameters(module: nn.Module) -> int:
    """Returns the total number of parameters in the model, including non-trainable."""
    return sum(p.numel() for p in module.parameters())


def get_named_modules(module: nn.Module, **kwargs: Any) -> List[Dict[str, str]]:
    """Obtain a list of named modules in the model/module.

    Return an iterator over all modules in the network, yielding both the name
    of the module as well as the module itself.

    Parameters
    ----------
    module : nn.Module
        The model to extract named modules from.
    **kwargs : Any
        Additional keyword arguments to pass to the `named_modules` method.

    Returns
    -------
    List[Dict[str, str]]
        A list of dictionaries containing the name and type of each module in the model.

    Examples
    --------
    >>> from torchvision.models import resnet18, ResNet18_Weights
    >>> backbone = resnet18(weights=ResNet18_Weights.DEFAULT)
    >>> named_modules = get_named_modules(backbone)

    ```python
    some_module = nn.ModuleDict(
        {"backbone": nn.Sequential(
            nn.Embedding(2, 2),
        ),
        "pooler":
            nn.Sequential(
                nn.Linear(1, 2),
            ),
        "head": nn.Sequential(
            nn.Conv2d(1, 3, 1),
        )
    })
    ```

    The model definition is:

    ```python
    ModuleDict(
    (backbone): Sequential(
        (0): Embedding(2, 2)
    )
    (pooler): Sequential(
        (0): Linear(in_features=1, out_features=2, bias=True)
    )
    (head): Sequential(
        (0): Conv2d(1, 3, kernel_size=(1, 1), stride=(1, 1))
    )
    )
    ```

    Then if we call `get_named_modules(some_module)`, we will get the following output:
    ```python
    [
        {
            "": "ModuleDict(backbone): Sequential((0): Embedding(2, 2))pooler): Sequential((0): Linear(in_features=1, out_features=2, bias=True))head): Sequential((0): Conv2d(1, 3, kernel_size=(1, 1), stride=(1, 1)))\n)"
        },
        {"backbone": "Sequential(0): Embedding(2, 2)\n)"},
        {"backbone.0": "Embedding(2, 2)"},
        {"pooler": "Sequential(0): Linear(in_features=1, out_features=2, bias=True)\n)"},
        {"pooler.0": "Linear(in_features=1, out_features=2, bias=True)"},
        {"head": "Sequential(0): Conv2d(1, 3, kernel_size=(1, 1), stride=(1, 1))\n)"},
        {"head.0": "Conv2d(1, 3, kernel_size=(1, 1), stride=(1, 1))"},
    ]
    ```
    """
    named_modules = []
    for _module in module.named_modules(**kwargs):
        module_name, module_type = _module
        named_modules.append({str(module_name): str(module_type)})
    return named_modules


def get_named_parameters(module: nn.Module, **kwargs: Any) -> List[Dict[str, nn.Parameter]]:
    """Obtain a list of named parameters in the model/module.

    Return an iterator over all parameters in the network, yielding both the name
    of the parameter as well as the parameter itself.

    Parameters
    ----------
    module : nn.Module
        The model to extract named parameters from.
    **kwargs : Any
        Additional keyword arguments to pass to the `named_parameters` method.

    Returns
    -------
    List[Dict[str, nn.Parameter]]
        A list of dictionaries containing the name and parameter of each parameter in the model.

    Examples
    --------
    >>> from torchvision.models import resnet18, ResNet18_Weights
    >>> backbone = resnet18(weights=ResNet18_Weights.DEFAULT)
    >>> named_parameters = get_named_parameters(backbone)

    ```python
    some_module = nn.ModuleDict(
        {"backbone": nn.Sequential(
            nn.Embedding(2, 2),
        ),
        "pooler":
            nn.Sequential(
                nn.Linear(1, 2),
            ),
        "head": nn.Sequential(
            nn.Conv2d(1, 3, 1),
        )
    })
    ```

    The model definition is:

    ```python
    ModuleDict(
    (backbone): Sequential(
        (0): Embedding(2, 2)
    )
    (pooler): Sequential(
        (0): Linear(in_features=1, out_features=2, bias=True)
    )
    (head): Sequential(
        (0): Conv2d(1, 3, kernel_size=(1, 1), stride=(1, 1))
    )
    )
    ```

    Then if we call `get_named_parameters(some_module)`, we will get the following output:

    ```text
    [
        {
            'backbone.0.weight': Parameter containing:
            tensor([
                [ 0.4833, -0.9517],
                [ 0.7517, -0.3315]
            ], requires_grad=True)
        },
        {
            'pooler.0.weight': Parameter containing:
            tensor([
                [ 0.8330],
                [-0.0932]
            ], requires_grad=True)
        },
        {
            'pooler.0.bias': Parameter containing:
            tensor([
                -0.5987,
                0.2774
            ], requires_grad=True)
        },
        {
            'head.0.weight': Parameter containing:
            tensor([
                [[[-0.3349]]],
                [[[-0.4394]]],
                [[[ 0.2738]]]
            ], requires_grad=True)
        },
        {
            'head.0.bias': Parameter containing:
            tensor([
                0.5669,
                0.7181,
                -0.0820
            ], requires_grad=True)
        }
    ]
    ```
    """
    named_parameters = []
    for _parameter in module.named_parameters(**kwargs):
        parameter_name, parameter = _parameter
        named_parameters.append({str(parameter_name): parameter})
    return named_parameters


def gather_weight_stats(module: nn.Module, **kwargs: Any) -> Dict[str, Dict[str, float]]:
    """Return the mean and standard deviation of weights and biases in the model. Sanity
    check to ensure that the weights and biases are initialized correctly.

    Parameters
    ----------
    module : nn.Module
        The model to extract weight statistics from.
    **kwargs : Any
        Additional keyword arguments to pass to the `named_modules` method.

    Returns
    -------
    Dict[str, Dict[str, float]]
        A dictionary containing the mean and standard deviation of weights and biases in the model.

    Examples
    --------
    >>> from torchvision.models import resnet18, ResNet18_Weights
    >>> backbone = resnet18(weights=ResNet18_Weights.DEFAULT)
    >>> stats = gather_weight_stats(backbone)
    >>> stats_df = prepare_stats_dataframe(stats)
    >>> plot_distribution_stats(stats_df)
    >>> plot_distribution_stats(stats_df, layer_type="Conv2d")
    """
    stats = {}
    for _module in module.named_modules(**kwargs):
        module_name, module_type = _module
        assert isinstance(module_type, nn.Module)
        if (
            hasattr(module_type, "weight")
            and isinstance(module_type.weight, torch.nn.Parameter)
            and module_type.weight is not None
        ):
            weight = module_type.weight.data
            weight_key = f"{module_name}+{str(module_type.__class__.__name__)}_weight"
            stats[weight_key] = {"w_mean": weight.mean().item(), "w_std": weight.std().item()}

        if (
            hasattr(module_type, "bias")
            and isinstance(module_type.bias, torch.nn.Parameter)
            and module_type.bias is not None
        ):
            bias = module_type.bias.data
            bias_key = f"{module_name}+{str(module_type.__class__.__name__)}_bias"
            stats[bias_key] = {"b_mean": bias.mean().item(), "b_std": bias.std().item()}
    return stats


def prepare_stats_dataframe(stats_dict: Dict[str, Dict[str, float]]) -> pd.DataFrame:
    """Use in conjunction with the `gather_weight_stats` function to prepare a DataFrame
    for visualization of weight and bias statistics."""
    data = []
    for key, values in stats_dict.items():
        layer_name, layer_type = key.rsplit("+", 1)
        layer_name = "+".join(layer_name.split("+")[:-1])
        layer_type = layer_type.replace("_weight", "").replace("_bias", "")

        data.append(
            {
                "layer_name": layer_name,
                "layer_type": layer_type,
                "parameter_type": "Weight" if "weight" in key else "Bias",
                "mean": values["w_mean"] if "weight" in key else values["b_mean"],
                "std": values["w_std"] if "weight" in key else values["b_std"],
            }
        )
    return pd.DataFrame(data)


def plot_distribution_stats(df: pd.DataFrame, layer_type: str | None = None) -> None:
    """Use in conjunction with the `prepare_stats_dataframe` function to visualize the
    distribution of weight and bias statistics."""
    if layer_type:
        df = df[df["layer_type"] == layer_type]

    fig, axes = plt.subplots(2, 2, figsize=(14, 12))
    sns.histplot(df[df["parameter_type"] == "Weight"], x="mean", kde=True, color="blue", ax=axes[0, 0])
    axes[0, 0].set_title("Distribution of Weights Means")
    sns.histplot(df[df["parameter_type"] == "Weight"], x="std", kde=True, color="blue", ax=axes[0, 1])
    axes[0, 1].set_title("Distribution of Weights Standard Deviations")

    sns.histplot(df[df["parameter_type"] == "Bias"], x="mean", kde=True, color="red", ax=axes[1, 0])
    axes[1, 0].set_title("Distribution of Biases Means")
    sns.histplot(df[df["parameter_type"] == "Bias"], x="std", kde=True, color="red", ax=axes[1, 1])
    axes[1, 1].set_title("Distribution of Biases Standard Deviations")

    plt.tight_layout()
    plt.show()


def compare_models(model_a: nn.Module, model_b: nn.Module) -> bool:
    """
    Compare two PyTorch models to check if they have identical parameters.

    Parameters
    ----------
    model_a : nn.Module
        The first model to compare.
    model_b : nn.Module
        The second model to compare.

    Returns
    -------
    bool
        Returns True if both models have identical parameters, False otherwise.
    """
    return all(
        torch.equal(param_a[1], param_b[1])
        for param_a, param_b in zip(model_a.state_dict().items(), model_b.state_dict().items())
    )


def compare_models_and_report_differences(model_a: nn.Module, model_b: nn.Module) -> Tuple[bool, Any]:
    """
    Compare two PyTorch models to check if they have identical parameters.

    Parameters
    ----------
    model_a : nn.Module
        The first model to compare.
    model_b : nn.Module
        The second model to compare.

    Returns
    -------
    Tuple[bool, Any]
        Returns a tuple with the first element as a boolean indicating if the
        models are identical. The second element is a dictionary containing the
        differences between the models if they are not identical.
    """
    model_a_dict = model_a.state_dict()
    model_b_dict = model_b.state_dict()

    if set(model_a_dict.keys()) != set(model_b_dict.keys()):
        # Early exit if model architectures are different (different sets of parameter keys)
        return False, {"error": "Models have different architectures and cannot be compared."}

    differences = {}
    for name in model_a_dict.keys():  # noqa: SIM118
        param_a = model_a_dict[name]
        param_b = model_b_dict[name]
        if not torch.equal(param_a, param_b):
            differences[name] = {
                "model_a": param_a.detach().cpu().numpy(),
                "model_b": param_b.detach().cpu().numpy(),
            }

    if differences:
        return False, differences
    else:
        return True, None


def _init_weights(module: nn.Module, **kwargs: Any) -> None:
    """Initialize the weights. This is a helper function to initialize the weights
    of a torch model.

    Parameters
    ----------
    module : nn.Module
        The module to initialize the weights. It can be the model itself, or
        a sub-module of the model such as `nn.Linear`, `nn.Embedding`, `nn.LayerNorm`.
    kwargs : Any
        Additional keyword arguments to pass to the initialization method.
        This is like a _universal_ way to accept configuration.

    Examples
    --------
    >>> from torch import nn
    >>> model = nn.Sequential(
    ...     nn.Linear(10, 10),
    ...     nn.ReLU(),
    ...     nn.Embedding(10, 10),
    ...     nn.LayerNorm(10),
    ... )
    >>> model.apply(lambda module: _init_weights(module, init_method="xavier_uniform", gain=10.0))
    """
    # std = self.config.initializer_range
    init_method = kwargs.get("init_method", "normal")

    if isinstance(module, nn.Linear):
        if init_method == "normal":
            module.weight.data.normal_(mean=kwargs.get("mean", 0.0), std=kwargs.get("std", 0.02))
        elif init_method == "xavier_uniform":
            module.weight.data = nn.init.xavier_uniform_(module.weight.data, gain=kwargs.get("gain", 1.0))
        elif init_method == "xavier_normal":
            module.weight.data = nn.init.xavier_normal_(module.weight.data, gain=kwargs.get("gain", 1.0))
        elif init_method == "kaiming_uniform":
            module.weight.data = nn.init.kaiming_uniform_(
                module.weight.data,
                kwargs.get("a", 0),
                kwargs.get("mode", "fan_in"),
                kwargs.get("nonlinearity", "leaky_relu"),
            )
        elif init_method == "kaiming_normal":
            module.weight.data = nn.init.kaiming_normal_(
                module.weight.data,
                kwargs.get("a", 0),
                kwargs.get("mode", "fan_in"),
                kwargs.get("nonlinearity", "leaky_relu"),
            )
        elif init_method == "orthogonal":
            module.weight.data = nn.init.orthogonal_(module.weight.data, kwargs.get("gain", 1.0))

        if module.bias is not None:
            module.bias.data.zero_()

    elif isinstance(module, nn.Embedding):
        if init_method == "normal":
            module.weight.data.normal_(mean=kwargs.get("mean", 0.0), std=kwargs.get("std", 0.02))
        elif init_method == "xavier_uniform":
            module.weight.data = nn.init.xavier_uniform_(module.weight.data, gain=kwargs.get("gain", 1.0))
        elif init_method == "xavier_normal":
            module.weight.data = nn.init.xavier_normal_(module.weight.data, gain=kwargs.get("gain", 1.0))
        elif init_method == "kaiming_uniform":
            module.weight.data = nn.init.kaiming_uniform_(
                module.weight.data,
                kwargs.get("a", 0),
                kwargs.get("mode", "fan_in"),
                kwargs.get("nonlinearity", "leaky_relu"),
            )
        elif init_method == "kaiming_normal":
            module.weight.data = nn.init.kaiming_normal_(
                module.weight.data,
                kwargs.get("a", 0),
                kwargs.get("mode", "fan_in"),
                kwargs.get("nonlinearity", "leaky_relu"),
            )
        elif init_method == "orthogonal":
            module.weight.data = nn.init.orthogonal_(module.weight.data, kwargs.get("gain", 1.0))

        if module.padding_idx is not None:
            module.weight.data[module.padding_idx].zero_()

    elif isinstance(module, nn.LayerNorm):
        module.bias.data.zero_()
        module.weight.data.fill_(1.0)


class Freezer:
    """A utility class for freezing and unfreezing parameters in a PyTorch model.

    This class allows for freezing the parameters of a model's layers either by
    the index of the layer or by the name of the parameters. It keeps track of
    the freeze state of each parameter, allowing for selective training of model parts.

    Examples
    --------
    >>> from rich.pretty import pprint
    >>> from torch import nn
    >>> model = nn.Sequential(
    ...     nn.Conv2d(1, 20, 5, bias=False),
    ...     nn.Conv2d(20, 50, 5, bias=False),
    ...     nn.Linear(800, 500, bias=False),
    ...     nn.Linear(500, 10, bias=False),
    ... )
    >>> freezer = Freezer(model)
    >>> freezer.freeze_by_index([0, 2])
    >>> status = freezer.report_freezing()
    >>> pprint(status)
    {'0.weight': True, '1.weight': False, '2.weight': True, '3.weight': False}

    >>> assert model[0].weight.requires_grad is False
    >>> assert model[1].weight.requires_grad is True
    >>> assert model[2].weight.requires_grad is False
    >>> assert model[3].weight.requires_grad is True

    >>> freezer.freeze_by_name(["3.weight"])
    >>> status = freezer.report_freezing()
    >>> pprint(status)
    {'0.weight': True, '1.weight': False, '2.weight': True, '3.weight': True}
    """

    def __init__(self, model: nn.Module) -> None:
        """Initialize the Freezer with the model. Note given the nature of mutable
        objects, the `model` is a reference to the original model from outside the
        constructor, and any modifications to the model will be reflected inplace.
        """
        self.model = model
        self.frozen_status: Dict[str, bool] = {}
        self.update_freezing_status()

    def update_freezing_status(self) -> None:
        """
        Update the freezing status of all parameters in the model. So `True` means
        the parameter is frozen, and `False` means it is trainable.
        """
        for name, param in self.model.named_parameters():
            self.frozen_status[name] = not param.requires_grad

    def freeze_by_index(self, indices: List[int], submodule_path: str | None = None) -> None:
        """
        Freeze layers based on their index in the ordered list of model's children.

        Parameters
        ----------
        indices : List[int]
            List of indices of the layers to freeze.
        submodule_path : str, optional
            The path to the submodule to freeze in the model. For example, if the model
            has a submodule `backbone` with a Conv2d layer at index 0, then the path
            would be `backbone`.
        """

        target = self.model
        if submodule_path:
            for attr in submodule_path.split("."):
                target = getattr(target, attr)

        for idx, child in enumerate(target.children()):
            if idx in indices:
                for param in child.parameters():
                    param.requires_grad = False
        self.update_freezing_status()

    def freeze_by_name(self, names: List[str]) -> None:
        """
        Freeze layers based on the names of the parameters.

        Parameters
        ----------
        names : List[str]
            List of substrings to match in the parameter names for freezing.
            This name can be obtained by calling `model.named_parameters()`.
        """
        for parameter_name, parameter in self.model.named_parameters():
            if any(n in parameter_name for n in names):
                parameter.requires_grad = False
        self.update_freezing_status()

    def freeze_by_module(self, module: nn.Module) -> None:
        """
        Freeze all parameters in the specified module.

        Parameters
        ----------
        module : nn.Module
            The module to freeze the parameters in.
        """
        for param in module.parameters():
            param.requires_grad = False
        self.update_freezing_status()

    def report_freezing(self) -> Dict[str, bool]:
        """
        Report the freezing status of all parameters in the model.

        Returns
        -------
        Dict[str, bool]
            Dictionary with parameter names and their freezing status.
        """
        return self.frozen_status


def freeze_layers(module: nn.Module) -> None:
    """Freezes the specified number of layers in the backbone. See `Freezer` for
    a more comprehensive way to freeze layers in a model.

    Parameters
    ----------
    module : nn.Module
        The model to freeze the layers in.
    """
    for parameter in module.parameters():
        parameter.requires_grad = False


def check_optimizer_coverage(model: nn.Module, optimizer: optim.Optimizer) -> Dict[str, nn.Parameter]:
    """
    Checks if all parameters in the given model are covered by the optimizer's
    parameter groups.

    Parameters
    ----------
    model: nn.Module
        The model for which the optimizer is used.
    optimizer: optim.Optimizer
        The optimizer that should be checked.

    Returns
    -------
    uncovered_params: Dict[str, nn.Parameter]
        A dictionary containing all model parameters that are not covered by the
        optimizer's parameter groups. The keys are the parameter names and the
        values are the corresponding parameter tensors.
    """
    # 1. we gather all model parameters with names
    model_params: Dict[str, nn.Parameter] = dict(model.named_parameters())

    # 2. we gather all optimizer parameters
    optimizer_params: Set[nn.Parameter] = {param for group in optimizer.param_groups for param in group["params"]}

    # 3. we check if all parameters are covered
    uncovered_params: Dict[str, nn.Parameter] = {
        name: param for name, param in model_params.items() if param not in optimizer_params
    }
    return uncovered_params


class LayerInfo(TypedDict):
    layer_name: str
    layer_parameters: int
    dtype: str
    trainable: bool


def get_model_layer_info(model: nn.Module) -> List[LayerInfo]:
    """
    A utility function to get information about the layers in a PyTorch model.

    Parameters
    ----------
    model: nn.Module
        The model to check the dtype and trainable status.

    Returns
    -------
    List[LayerInfo]
        A list of dictionaries containing the layer name, number of parameters,
        dtype, and whether the layer is trainable or not.

    Examples
    --------
    >>> from transformers import AutoModelForSequenceClassification
    >>> model = AutoModelForSequenceClassification.from_pretrained("microsoft/deberta-v3-small")
    >>> model_layers_info = get_model_layer_info(model)
    """
    model_layers_info: List[LayerInfo] = []
    for parameter_name, parameter in model.named_parameters():
        this_layer_num_parameters = int(torch.prod(torch.tensor(parameter.size())))
        model_layers_info.append(
            {
                "layer_name": parameter_name,
                "layer_parameters": this_layer_num_parameters,
                "dtype": str(parameter.data.dtype),
                "trainable": bool(parameter.requires_grad),
            }
        )
    return model_layers_info



from __future__ import annotations

import inspect
from inspect import Parameter, Signature
from typing import Any, Callable, Dict, List, Set, Tuple, Type, get_type_hints, overload


@overload
def get_members_of_function_or_method(
    func_or_class: Type[object], predicate: Callable[[Any], bool] | None = None
) -> List[Tuple[str, Any]]:
    ...


@overload
def get_members_of_function_or_method(
    func_or_class: Callable[..., Any], predicate: Callable[[Any], bool] | None = None
) -> List[Tuple[str, Any]]:
    ...


def get_members_of_function_or_method(
    func_or_class: Type[object] | Callable[..., Any], predicate: Callable[[Any], bool] | None = None
) -> List[Tuple[str, Any]]:
    return inspect.getmembers(func_or_class, predicate)


def get_base_classes(cls: Type[Any], include_self: bool = False) -> Set[Type[Any]]:
    """
    Get the base classes of a class and all its base classes.
    """
    return set(cls.__mro__[0:-1] if include_self else cls.__mro__[1:-1])


def get_default(param: Parameter) -> Any:
    """Return the parameter's default value or None if not specified."""
    return param.default if param.default is not param.empty else None


def get_field_annotations(func_or_method: Callable[..., Any]) -> Tuple[List[Tuple[str, Any, Any]], Dict[str, Any]]:
    if not inspect.isroutine(func_or_method):
        raise ValueError("Expected a function or method")

    required_fields = []
    optional_fields = []
    annotations = {}

    try:
        sig: Signature = inspect.signature(func_or_method)
        type_hints: Dict[str, Any] = get_type_hints(func_or_method)
    except ValueError:
        raise ValueError("Object does not support signature or type hints extraction.") from None

    for name, param in sig.parameters.items():
        if name == "self":
            continue

        type_hint = type_hints.get(name, Any)
        annotations[name] = type_hint
        if param.default is param.empty:
            required_fields.append((name, type_hint, Ellipsis))
        else:
            default_value = param.default
            optional_fields.append((name, type_hint, default_value))

    fields = required_fields + optional_fields
    return fields, annotations


# TODO: Tuple[str, str, Any, Any] should be Tuple[str, str, Any, ellipsis]
def get_constructor_field_annotations(
    cls: Type[Any], include_bases: bool = True
) -> Tuple[List[Tuple[str, str, Any, Any]], Dict[str, Any]]:
    fields = []
    annotations = {}

    classes_to_inspect = [cls] + list(get_base_classes(cls, include_self=False)) if include_bases else [cls]

    for c in reversed(classes_to_inspect):  # Reverse to respect MRO
        if hasattr(c, "__init__"):
            class_fields, class_annotations = get_field_annotations(c.__init__)
            # Update fields and annotations with those from the current class,
            # avoiding duplicates.
            class_name = f"{c.__module__}.{c.__name__}"
            annotations[class_name] = class_annotations

            for name, type_hint, default in class_fields:
                fields.append((class_name, name, type_hint, default))

    return fields, annotations



"""Base class for vector plotter."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Generic, Optional

from omnivault._types._generic import Vec
from omnivault.utils.visualization.figure_manager import FigureManager


class VectorPlotter(Generic[Vec], FigureManager, ABC):
    @abstractmethod
    def plot(self, grid: bool = True, show_ticks: bool = False) -> None:
        ...

    @abstractmethod
    def add_vector(self, vector: Vec) -> None:
        ...

    @abstractmethod
    def add_text(
        self,
        x: float,
        y: float,
        text: str,
        z: Optional[float] = None,
        fontsize: int = 16,
        **kwargs: Any,
    ) -> None:
        ...



"""See `transformers/optimization.py` for the original code. This is a distilled
version."""
from __future__ import annotations

import math
from enum import Enum
from typing import Any, Dict, Type

import torch
from transformers import (
    get_constant_schedule,
    get_constant_schedule_with_warmup,
    get_cosine_schedule_with_warmup,
    get_cosine_with_hard_restarts_schedule_with_warmup,
    get_inverse_sqrt_schedule,
    get_linear_schedule_with_warmup,
    get_polynomial_decay_schedule_with_warmup,
)


class ExplicitEnum(str, Enum):
    """
    Enum with more explicit error message for missing values.
    """

    @classmethod
    def _missing_(cls: Type[Enum], value: object) -> Enum:
        raise ValueError(
            f"{value} is not a valid {cls.__name__}, please select one of {list(cls._value2member_map_.keys())}"
        )


class SchedulerType(ExplicitEnum):
    LINEAR = "linear"
    COSINE = "cosine"
    COSINE_WITH_RESTARTS = "cosine_with_restarts"
    POLYNOMIAL = "polynomial"
    CONSTANT = "constant"
    CONSTANT_WITH_WARMUP = "constant_with_warmup"
    INVERSE_SQRT = "inverse_sqrt"
    REDUCE_ON_PLATEAU = "reduce_lr_on_plateau"
    COSINE_WITH_MIN_LR = "cosine_with_min_lr"
    WARMUP_STABLE_DECAY = "warmup_stable_decay"


# transformers/optimization.py
TYPE_TO_SCHEDULER_FUNCTION: Dict[SchedulerType, Type[torch.optim.lr_scheduler._LRScheduler]] = {
    SchedulerType.LINEAR: get_linear_schedule_with_warmup,
    SchedulerType.COSINE: get_cosine_schedule_with_warmup,
    SchedulerType.COSINE_WITH_RESTARTS: get_cosine_with_hard_restarts_schedule_with_warmup,
    SchedulerType.POLYNOMIAL: get_polynomial_decay_schedule_with_warmup,
    SchedulerType.CONSTANT: get_constant_schedule,
    SchedulerType.CONSTANT_WITH_WARMUP: get_constant_schedule_with_warmup,
    SchedulerType.INVERSE_SQRT: get_inverse_sqrt_schedule,
}


def get_warmup_steps_from_ratio(total_train_steps: int, warmup_ratio: float) -> int:
    """
    Get number of steps used for a linear warmup.
    """
    return math.ceil(total_train_steps * warmup_ratio)


def get_scheduler(
    name: str | SchedulerType,
    optimizer: torch.optim.Optimizer,
    num_warmup_steps: int | None = None,
    num_training_steps: int | None = None,
    scheduler_specific_kwargs: Dict[str, Any] | None = None,
) -> torch.optim.lr_scheduler._LRScheduler:
    """
    Unified API to get any scheduler from its name.

    Args:
        name (`str` or `SchedulerType`):
            The name of the scheduler to use.
        optimizer (`torch.optim.Optimizer`):
            The optimizer that will be used during training.
        num_warmup_steps (`int`, *optional*):
            The number of warmup steps to do. This is not required by all schedulers (hence the argument being
            optional), the function will raise an error if it's unset and the scheduler type requires it.
        num_training_steps (`int``, *optional*):
            The number of training steps to do. This is not required by all schedulers (hence the argument being
            optional), the function will raise an error if it's unset and the scheduler type requires it.
        scheduler_specific_kwargs (`dict`, *optional*):
            Extra parameters for schedulers such as cosine with restarts. Mismatched scheduler types and scheduler
            parameters will cause the scheduler function to raise a TypeError.
    """
    name = SchedulerType(name)
    schedule_func = TYPE_TO_SCHEDULER_FUNCTION[name]

    if name == SchedulerType.CONSTANT:
        return schedule_func(optimizer)

    if scheduler_specific_kwargs is None:
        scheduler_specific_kwargs = {}

    # All other schedulers require `num_warmup_steps`
    if num_warmup_steps is None:
        raise ValueError(f"{name} requires `num_warmup_steps`, please provide that argument.")

    if name == SchedulerType.CONSTANT_WITH_WARMUP:
        return schedule_func(optimizer, num_warmup_steps=num_warmup_steps)  # type: ignore[call-arg]

    if name == SchedulerType.INVERSE_SQRT:
        return schedule_func(optimizer, num_warmup_steps=num_warmup_steps)  # type: ignore[call-arg]

    # All other schedulers require `num_training_steps`
    if num_training_steps is None:
        raise ValueError(f"{name} requires `num_training_steps`, please provide that argument.")

    return schedule_func(  # type: ignore[call-arg]
        optimizer,
        num_warmup_steps=num_warmup_steps,
        num_training_steps=num_training_steps,
        **scheduler_specific_kwargs,
    )



# NOTE: unit-tested
"""Utilities to evaluate the clustering performance of models."""
from __future__ import annotations

from typing import Any, Literal, overload

import numpy as np
import pandas as pd
from numpy.typing import NDArray


def contingency_matrix(
    y_trues: NDArray[np.integer[Any]], y_preds: NDArray[np.integer[Any]], as_dataframe: bool = False
) -> pd.DataFrame | NDArray[np.integer[Any]]:
    """
    Generate a contingency matrix for clustering, analogous to a confusion matrix in
    classification.

    This matrix counts the number of times each unique element of the true dataset
    (rows) occurs together with each unique element of the predicted dataset
    (columns). This function is particularly useful for evaluating clustering by
    comparing the true labels and predicted labels.

    Parameters
    ----------
    y_trues : NDArray[np.integer[Any]]
        An array of ground truth (true) labels.
    y_preds : NDArray[np.integer[Any]]
        An array of predicted labels, typically from clustering.
    as_dataframe : bool, optional
        If True, returns the contingency matrix as a pandas DataFrame with labeled rows and columns;
        otherwise, returns it as a NumPy array.

    Returns
    -------
    Union[pd.DataFrame, NDArray[np.integer[Any]]]
        The contingency matrix as either a pandas DataFrame or a NumPy array, depending on the
        value of `as_dataframe`.

    Notes
    -----
    The contingency matrix does not ensure that each predicted cluster label is
    assigned only once to a true label. This might lead to potential ambiguities in
    cluster labeling, especially when the number of predicted clusters does not
    match the number of true classes.

    Examples
    --------
    >>> y_trues = np.array([1, 2, 1, 2, 1])
    >>> y_preds = np.array([0, 0, 1, 1, 1])
    >>> print(contingency_matrix(y_trues, y_preds))
    [[0 3]
     [2 1]]

    >>> print(contingency_matrix(y_trues, y_preds, as_dataframe=True))
           pred=0  pred=1
    true=1      0       3
    true=2      2       1
    """
    # fmt: off
    classes, class_idx        = np.unique(y_trues, return_inverse=True)     # get the unique classes and their indices
    clusters, cluster_idx     = np.unique(y_preds, return_inverse=True)  # get the unique clusters and their indices
    num_classes, num_clusters = classes.shape[0], clusters.shape[0]  # get the number of classes and clusters
    # fmt: on

    # initialize the contingency matrix with shape num_classes x num_clusters
    # exactly the same as the confusion matrix but in confusion matrix
    # the rows are the true labels and the columns are the predicted labels
    # and hence is num_classes x num_classes instead of num_classes x num_clusters
    # however in kmeans for example it is possible to have len(np.unique(y_true)) != len(np.unique(y_pred)
    # i.e. the number of clusters is not equal to the number of classes
    contingency_matrix = np.zeros((num_classes, num_clusters), dtype=np.int64)

    # note however y_true and y_pred are same sequence of samples
    for i in range(class_idx.shape[0]):
        # loop through each sample and increment the contingency matrix
        # at the row corresponding to the true label and column corresponding to the predicted label
        # so if the sample index is i = 0, and class_idx[i] = 1 and cluster_idx[i] = 2
        # this means the gt label is 1 and the predicted label is 2
        # so we increment the contingency matrix at row 1 and column 2 by 1
        # then for each row, which is the row for each gt label,
        # we see which cluster has the highest number of samples and that is the cluster
        # that the gt label is most likely to belong to.
        # in other words since kmeans permutes the labels, we can't just compare the labels
        # directly.
        contingency_matrix[class_idx[i], cluster_idx[i]] += 1

    # row is the true label and column is the predicted label
    if as_dataframe:
        return pd.DataFrame(
            contingency_matrix,
            index=[f"true={c}" for c in classes],
            columns=[f"pred={c}" for c in clusters],
        )
    return contingency_matrix


@overload
def purity_score(
    y_trues: NDArray[np.integer[Any]], y_preds: NDArray[np.integer[Any]], per_cluster: Literal[True]
) -> NDArray[np.floating[Any]]:
    ...


@overload
def purity_score(
    y_trues: NDArray[np.integer[Any]], y_preds: NDArray[np.integer[Any]], per_cluster: Literal[False] = False
) -> float:
    ...


def purity_score(
    y_trues: NDArray[np.integer[Any]], y_preds: NDArray[np.integer[Any]], per_cluster: bool = False
) -> float | NDArray[np.floating[Any]]:
    """
    Compute the purity score for clustering evaluation.

    Purity is an external evaluation metric of clustering
    quality. It is the measure of the largest fraction of
    true labels in each cluster. The overall purity is the
    sum of purities of all clusters divided by the total
    number of samples.

    Parameters
    ----------
    y_trues : NDArray[np.integer[Any]]
        Array of ground truth (true) labels.
    y_preds : NDArray[np.integer[Any]]
        Array of predicted labels from clustering.
    per_cluster : bool, optional
        If True, returns the purity for each cluster
        instead of the overall purity score.

    Returns
    -------
    Union[float, NDArray[np.floating[Any]]]
        If `per_cluster` is False, returns a single float
        value representing the overall purity of the
        clustering. If True, returns an array of purity
        scores for each cluster.

    Notes
    -----
    - Purity can be misleading in cases where cluster size
      varies significantly. A large cluster that is only
      moderately pure can disproportionately influence the
      overall purity score.
    - Purity does not penalize having many clusters, so
      it should not be used in isolation.
    """
    # compute contingency matrix (also called confusion matrix)
    contingency_matrix_ = contingency_matrix(y_trues, y_preds, as_dataframe=False)

    # total purity is the max value in each column divided by the sum of the matrix
    # this means for each cluster k, we find the gt label that has the most samples in that cluster
    # and then sum up all clusters and we divide by the total number of samples in all clusters

    # if per_cluster is True, we return the purity for each cluster
    # this means instead of sum up all clusters, we return the purity for each cluster.
    if per_cluster:
        return np.amax(contingency_matrix_, axis=0) / np.sum(contingency_matrix_, axis=0)  # type: ignore[no-any-return]
    # return purity which is the sum of the max values in each column divided by the sum of the matrix
    return float(np.sum(np.amax(contingency_matrix_, axis=0)) / np.sum(contingency_matrix_))



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generator, Generic, Iterable, List, Optional, overload

from omnivault._types._generic import T


class Stack(ABC, Generic[T]):
    """
    This interface defines the contract for a stack data structure.
    """

    @overload
    def __init__(self) -> None:
        ...

    @overload
    def __init__(self, iterable: Iterable[T]) -> None:
        ...

    def __init__(self, iterable: Optional[Iterable[T]] = None) -> None:
        """Construct a new stack object.

        Parameters
        ----------
        iterable : Optional[Iterable[T]], optional
            An iterable to initialize the stack with, by default None
        """
        self._stack_items: List[T] = []
        if iterable is not None:
            for item in iterable:
                self.push(item)

    @abstractmethod
    def push(self, item: T) -> None:
        """Push an item on top of the stack."""
        raise NotImplementedError

    @abstractmethod
    def pop(self) -> T:
        """Pop an item from the top of the stack."""
        raise NotImplementedError

    @abstractmethod
    def peek(self) -> T:
        """Return the top most item in the stack without modifying the stack."""
        raise NotImplementedError

    @abstractmethod
    def is_empty(self) -> bool:
        """Check if stack is empty."""
        raise NotImplementedError

    @abstractmethod
    def __len__(self) -> int:
        """Return the size of the stack."""
        raise NotImplementedError

    @abstractmethod
    def __iter__(self) -> Generator[T, None, None]:
        """Return an iterator for the stack."""
        raise NotImplementedError



from __future__ import annotations

import logging
import sys
import time

from hydra.utils import instantiate
from omegaconf import DictConfig, ListConfig
from omegaconf import OmegaConf as om
from torch.utils.data import Subset

from omnivault._types._alias import Missing
from omnivault._types._sentinel import MISSING
from omnivault.core.logger import RichLogger
from omnivault.transformer.config.composer import Composer, DataConfig
from omnivault.transformer.config.constants import MaybeConstant
from omnivault.transformer.config.criterion import CRITERION_REGISTRY
from omnivault.transformer.config.decoder import DecoderConfig
from omnivault.transformer.config.generator import GeneratorConfig
from omnivault.transformer.config.global_ import MaybeGlobal
from omnivault.transformer.config.logger import LoggerConfig
from omnivault.transformer.config.optim import OPTIMIZER_REGISTRY
from omnivault.transformer.config.scheduler import SCHEDULER_REGISTRY
from omnivault.transformer.config.trainer import TrainerConfig
from omnivault.transformer.core.callbacks import save_state
from omnivault.transformer.core.dataset import TextCharacterDataset, create_loader, split_dataset
from omnivault.transformer.core.optim import apply_weight_decay_to_different_param_groups
from omnivault.transformer.core.state import State
from omnivault.transformer.core.tokenizer import TextCharacterTokenizer
from omnivault.transformer.core.trainer import Trainer, TrainerEvent
from omnivault.transformer.core.vocabulary import TextCharacterVocabulary
from omnivault.transformer.decoder.core import GPTDecoder
from omnivault.transformer.projects.tinyshakespeare_char.callbacks import evaluate_generate_on_train_batch_end
from omnivault.transformer.utils.visualization import save_plot_history
from omnivault.utils.config_management.omegaconf import load_yaml_config, merge_configs
from omnivault.utils.reproducibility.seed import seed_all


def main(cfg: DictConfig | ListConfig) -> None:
    """Main driver."""
    seed_all(cfg.global_.seed)

    constants = MaybeConstant(**cfg.constants) if cfg.constants is not None else MaybeConstant()
    logger_pydantic_config = LoggerConfig(**cfg.logger)
    global_ = MaybeGlobal(**cfg.global_)
    data = DataConfig(**cfg.data)
    trainer_config = TrainerConfig(**cfg.trainer)
    generator_config = GeneratorConfig(**cfg.generator)

    # logger
    logger = RichLogger(**logger_pydantic_config.model_dump(mode="python")).logger
    assert isinstance(logger, logging.Logger)

    vocabulary = TextCharacterVocabulary.from_url(
        url=data.dataset_url, dataset_name=data.dataset_name, dest_folder=data.dataset_dir
    )
    tokenizer = TextCharacterTokenizer(vocabulary=vocabulary)

    # assign back model.vocab_size from ??? to vocabulary.vocab_size
    cfg.model.vocab_size = vocabulary.vocab_size
    model_config = instantiate(cfg.model)
    model_pydantic_config = DecoderConfig(**model_config)

    optimizer_config_cls = OPTIMIZER_REGISTRY[cfg.optimizer.name]
    optimizer_pydantic_config = optimizer_config_cls(**cfg.optimizer)

    criterion_config_cls = CRITERION_REGISTRY[cfg.criterion.name]
    criterion_pydantic_config = criterion_config_cls(**cfg.criterion)

    composer = Composer(
        constants=constants,
        global_=global_,
        data=data,
        model=model_pydantic_config,
        optimizer=optimizer_pydantic_config,
        criterion=criterion_pydantic_config,
        trainer=trainer_config,
        generator=generator_config,
    )
    assert composer.model is not MISSING and not isinstance(composer.model, Missing)
    assert composer.optimizer is not MISSING and not isinstance(composer.optimizer, Missing)
    assert composer.criterion is not MISSING and not isinstance(composer.criterion, Missing)

    # TODO: consider classmethod from file_path
    assert composer.data.dataset_path is not None
    with open(composer.data.dataset_path, "r") as file:
        corpus = file.read()

    dataset = TextCharacterDataset(corpus=corpus, context_length=composer.data.context_length, tokenizer=tokenizer)

    if composer.global_.debug:
        debug_samples = composer.global_.debug_samples or 256
        dataset = Subset(dataset, indices=range(debug_samples))  # type: ignore[assignment]

    if composer.data.split:
        train_dataset, valid_dataset, test_dataset = split_dataset(
            dataset=dataset, split=composer.data.split, seed=composer.global_.seed
        )
    else:
        train_dataset = dataset  # type: ignore[assignment]
        valid_dataset = None
        test_dataset = None

    assert composer.data.train_loader is not None
    assert composer.data.collate_fn is not None

    train_loader = create_loader(
        dataset=train_dataset,
        loader_config=composer.data.train_loader,
        collate_fn_config=composer.data.collate_fn,
    )

    if valid_dataset is not None:
        assert composer.data.valid_loader is not None
        valid_loader = create_loader(  # noqa: F841
            dataset=valid_dataset,
            loader_config=composer.data.valid_loader,
            collate_fn_config=composer.data.collate_fn,
        )

    if test_dataset is not None:
        assert composer.data.test_loader is not None
        test_loader = create_loader(  # noqa: F841
            dataset=test_dataset,
            loader_config=composer.data.test_loader,
            collate_fn_config=composer.data.collate_fn,
        )

    # Create model
    model = GPTDecoder(model_pydantic_config)
    model = model.to(device=composer.trainer.device, dtype=next(model.parameters()).dtype, non_blocking=True)

    # Create optimizer based on model parameters
    if composer.trainer.apply_weight_decay_to_different_param_groups:
        assert hasattr(composer.optimizer, "weight_decay")
        optimizer = optimizer_pydantic_config.build(
            params=apply_weight_decay_to_different_param_groups(
                model=model, weight_decay=composer.optimizer.weight_decay
            )
        )
    else:
        optimizer = optimizer_pydantic_config.build(params=model.parameters())

    # Create criterion
    criterion = criterion_pydantic_config.create_instance()
    assert criterion.ignore_index == -100

    # Create scheduler
    scheduler_config_cls = SCHEDULER_REGISTRY[cfg.scheduler.name]
    scheduler_pydantic_config = scheduler_config_cls(**cfg.scheduler)  # type: ignore[assignment]
    assert composer.scheduler is MISSING  # now it is MISSING for us to fill up.
    composer.scheduler = scheduler_pydantic_config
    scheduler = scheduler_pydantic_config.build(optimizer=optimizer)

    composer.pretty_print()
    time.sleep(1)

    state = State(
        model=model,
        criterion=criterion,
        optimizer=optimizer,
        scheduler=scheduler,
        vocabulary=vocabulary,
        tokenizer=tokenizer,
    )
    state.pretty_print()
    time.sleep(1)

    device = composer.trainer.device

    # train
    trainer = Trainer(
        state=state,
        composer=composer,
        logger=logger,
        device=device,  # type: ignore[arg-type]
    )
    trainer.add_callback(TrainerEvent.ON_TRAIN_BATCH_END, evaluate_generate_on_train_batch_end)
    trainer.add_callback(TrainerEvent.ON_TRAIN_EPOCH_END, save_state)

    _trained_state = trainer.fit(train_loader=train_loader)
    history = _trained_state.history
    _ = save_plot_history(history, plot=False, save_path=f"{composer.trainer.save_dir}/history.png")


if __name__ == "__main__":
    yaml_path = sys.argv[1]
    args_list = sys.argv[2:]

    yaml_cfg = load_yaml_config(yaml_path)
    cfg = merge_configs(yaml_cfg, args_list)
    om.resolve(cfg)  # inplace ops

    main(cfg)



from __future__ import annotations

import re
from abc import ABC, abstractmethod
from typing import Generic, List, TypeVar, Union

import torch

from omnivault.transformer.core.vocabulary import AdderVocabulary, TextCharacterVocabulary, Vocabulary

Vocabulary_t = TypeVar("Vocabulary_t", bound=Vocabulary)


class Tokenizer(ABC, Generic[Vocabulary_t]):
    def __init__(self, vocabulary: Vocabulary):
        self.vocabulary = vocabulary

    # NOTE: for simplicity, I do not add more args in tokenizer, encode and decode
    # here in fear of violating the Liskov Substitution Principle, which is not
    # something I want to worry about in a naive implementation.

    @abstractmethod
    def tokenize(self, sequence: str, add_special_tokens: bool = True) -> List[str]:
        """
        Tokenizes a sequence into a sequence (list) of tokens.

        Example
        -------
        >>> tokenizer = Tokenizer()
        >>> tokenizer.tokenize("The quick brown fox jumps over the lazy dog.")
        ["The", "quick", "brown", "fox", "jumps", "over", "the", "lazy", "dog", "."]

        Parameters
        ----------
        sequence : str
            The sequence to tokenize.
        add_special_tokens : bool
            Whether to add special tokens to the sequence of tokens, by default True.

        Returns
        -------
        List[str]
            The sequence of tokens.
        """

    @abstractmethod
    def encode(self, sequence: str, add_special_tokens: bool = True) -> List[int]:
        """
        Encodes a sequence to its corresponding integer index.

        Parameters
        ----------
        sequence : str
            The sequence to encode.

        Returns
        -------
        List[int]
            The integer index corresponding to the token.
        """

    @abstractmethod
    def decode(self, encoded_sequence: List[int] | torch.Tensor, remove_special_tokens: bool = True) -> str:
        """
        Decodes an integer index back to its corresponding token.

        Parameters
        ----------
        index : int
            The integer index to decode.

        Returns
        -------
        str
            The token corresponding to the integer index.
        """


class AdderTokenizer(Tokenizer[AdderVocabulary]):
    def __init__(self, vocabulary: AdderVocabulary) -> None:
        super().__init__(vocabulary)

    def tokenize(self, sequence: str, add_special_tokens: bool = True) -> List[str]:
        tokens = [char for char in sequence]  # noqa: C416
        if add_special_tokens:
            tokens = [AdderVocabulary.BOS] + tokens + [AdderVocabulary.EOS]
        return tokens

    def encode(self, sequence: str, add_special_tokens: bool = True) -> List[int]:
        tokens: List[str] = self.tokenize(sequence, add_special_tokens=add_special_tokens)
        encoded_sequence: List[int] = [
            self.vocabulary.token_to_index.get(token, self.vocabulary.token_to_index[AdderVocabulary.UNK])
            for token in tokens
        ]
        return encoded_sequence

    def encode_batch(self, sequences: List[str], add_special_tokens: bool = True) -> List[List[int]]:
        return [self.encode(sequence, add_special_tokens=add_special_tokens) for sequence in sequences]

    def decode(self, encoded_sequence: List[int] | torch.Tensor, remove_special_tokens: bool = True) -> str:
        if isinstance(encoded_sequence, torch.Tensor):
            encoded_sequence = encoded_sequence.tolist()
        decoded = "".join([self.vocabulary.index_to_token.get(char, AdderVocabulary.UNK) for char in encoded_sequence])

        if remove_special_tokens:
            decoded = re.sub(
                f"{AdderVocabulary.BOS}|{AdderVocabulary.EOS}|{AdderVocabulary.PAD}|{AdderVocabulary.UNK}",
                "",
                decoded,
            )
        return decoded

    def decode_batch(
        self, encoded_sequences: List[List[int]] | torch.Tensor, remove_special_tokens: bool = True
    ) -> List[str]:
        return [
            self.decode(encoded_sequence, remove_special_tokens=remove_special_tokens)
            for encoded_sequence in encoded_sequences
        ]


class TextCharacterTokenizer(Tokenizer[TextCharacterVocabulary]):
    """
    A tokenizer for character-level text processing, responsible for tokenizing,
    encoding, and decoding text sequences using a given character vocabulary.
    """

    def tokenize(self, sequence: str, add_special_tokens: bool = False) -> List[str]:
        tokens = list(sequence)  # Tokenizes the text into a list of characters
        if add_special_tokens:
            tokens = [TextCharacterVocabulary.BOS] + tokens + [TextCharacterVocabulary.EOS]
        return tokens

    def encode(self, sequence: str, add_special_tokens: bool = False) -> List[int]:
        tokens = self.tokenize(sequence, add_special_tokens=add_special_tokens)
        return [self.vocabulary.token_to_index.get(char, -1) for char in tokens]  # -1 for unknown characters

    def decode(self, encoded_sequence: List[int] | torch.Tensor, remove_special_tokens: bool = False) -> str:
        if isinstance(encoded_sequence, torch.Tensor):
            encoded_sequence = encoded_sequence.tolist()

        decoded_sequence = "".join([self.vocabulary.index_to_token.get(char, "") for char in encoded_sequence])

        if remove_special_tokens:
            # Remove special tokens from the decoded string
            special_tokens = {self.vocabulary.BOS, self.vocabulary.EOS, self.vocabulary.PAD, self.vocabulary.UNK}
            decoded_sequence = "".join(char for char in decoded_sequence if char not in special_tokens)

        return decoded_sequence


Tokenizers = Union[AdderTokenizer, TextCharacterTokenizer]



from __future__ import annotations

import torch

__all__ = ["get_device"]


def get_device() -> torch.device:
    if torch.cuda.is_available():
        # auto select device
        device = f"cuda:{torch.cuda.current_device()}"
    elif torch.backends.mps.is_available():
        device = "mps"
    else:
        device = "cpu"
    return torch.device(device)



from abc import ABC, abstractmethod

import torch
from torch import nn


class PositionalEncoding(ABC, nn.Module):
    def __init__(self, d_model: int, context_length: int, dropout: float = 0.0) -> None:
        super().__init__()
        self.d_model = d_model
        self.context_length = context_length
        self.dropout = nn.Dropout(p=dropout, inplace=False)

    @abstractmethod
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        ...


class Sinusoid(PositionalEncoding):
    P: torch.Tensor

    def __init__(self, d_model: int, context_length: int, dropout: float = 0.0) -> None:
        super().__init__(d_model, context_length, dropout)

        P = self._init_positional_encoding()
        self.register_buffer("P", P, persistent=True)  # with this no need requires_grad=False

    def _init_positional_encoding(self) -> torch.Tensor:
        """Initialize the positional encoding tensor."""
        P = torch.zeros((1, self.context_length, self.d_model))
        position = self._get_position_vector()
        div_term = self._get_div_term_vector()
        P[:, :, 0::2] = torch.sin(position / div_term)
        P[:, :, 1::2] = torch.cos(position / div_term)
        return P

    def _get_position_vector(self) -> torch.Tensor:
        """Return a vector representing the position of each token in a sequence."""
        return torch.arange(self.context_length, dtype=torch.float32).reshape(-1, 1)

    def _get_div_term_vector(self) -> torch.Tensor:
        """Return a vector representing the divisor term for positional encoding."""
        return torch.pow(
            10000,
            torch.arange(0, self.d_model, 2, dtype=torch.float32) / self.d_model,
        )

    def forward(self, z: torch.Tensor) -> torch.Tensor:
        z = self._add_positional_encoding(z)
        z = self.dropout(z)
        return z

    def _add_positional_encoding(self, z: torch.Tensor) -> torch.Tensor:
        """Add the positional encoding tensor to the input tensor."""
        return z + self.P[:, : z.shape[1], :]



"""
## Reference

The proper implementation is in:
https://github.com/taleinat/python-stdlib-sentinels/blob/main/sentinels/sentinels.py
but it is slightly more complicated for this project.

## `NotGiven`

- **Purpose**: Indicates that a parameter was not provided at all. It's used to
  distinguish between a parameter being explicitly set to `None` and not being
  provided.
- **Use Case**: Common in APIs where default behavior is triggered when a
  parameter is not given, but `None` might be a valid, meaningful input. For
  example, `None` might mean "disable timeout", while `NotGiven` means "use a
  default timeout".
- Other example usage is if you want to assign a default empty list or dict but
  it is mutable, so you assign this type but not None since None don't make
  sense.
- **Behavior**: Functions can check for `NotGiven` to apply default behavior.

## `Omit`

- **Purpose**: Used to explicitly remove or omit a default value that would
  otherwise be applied. It's not just about a value being absent, but rather
  about actively removing a pre-existing default.
- **Use Case**: Useful in situations where the default behavior or value needs
  to be explicitly overridden or disabled, and where `None` is not a suitable
  option. For example, removing a default HTTP header.
- **Behavior**: Functions can check for `Omit` to actively remove or ignore a
  default setting or value.

### Comparison

- **Similarity**: Both are used to signal special cases in the absence of normal
  parameter values.
- **Difference**: `NotGiven` is about the absence of a value where a default may
  apply, while `Omit` is about actively overriding a default.
"""

from __future__ import annotations

import threading
from typing import Any, Dict, Generic, Literal, Type

from typing_extensions import override

from omnivault._types._generic import _T


class _NotGiven:
    """
    A sentinel singleton class used to distinguish omitted keyword arguments
    from those passed in with the value None (which may have different behavior).

    Quite similar with dataclass's MISSING.

    This is used to differentiate between cases where a parameter is not
    provided and where a parameter is provided with the value None. The class
    provides a more descriptive representation than None or other placeholders.

    NOTE: example usage is if you want to assign a default empty list or dict
    but it is mutable, so you assign this type but not None since None don't make
    sense.

    It is a singleton because `None` is also a singleton so we mimic this
    behaviour. No matter how many times you call `None` in any function or
    methods, it will reference the same unique singleton `None` class.

    More importantly, because `None` is a singleton, we can use the `is`
    operator to check for object identity. This is why the idiomatic way
    to check if a variable is `None` is to do `if var is None`.

    So that is why we make `_NotGiven` a singleton, because referencing this
    class across scripts will maintain its unique identity across imports.

    We further make this class immutable to behave a bit like `None`.

    Example
    -------
    ```python
    def get(timeout: Union[int, _NotGiven, None] = _NotGiven()) -> Response:
        if timeout is _NotGiven:
            # Default timeout behavior
        elif timeout is None:
            # No timeout
        else:
            # Specific timeout given

    get(timeout=1) # 1s timeout
    get(timeout=None) # No timeout
    get() # Default timeout behavior, which may not be statically known at
          # the method definition.
    ```
    """

    _instance: _NotGiven | None = None

    def __new__(cls: Type[_NotGiven]) -> _NotGiven:  # noqa: PYI034
        if cls._instance is None:
            cls._instance = super(_NotGiven, cls).__new__(cls)  # noqa: UP008
        return cls._instance

    def __bool__(self) -> Literal[False]:
        """
        This method is used to define the boolean value of an instance of `_NotGiven`.
        By returning `False`, it allows `_NotGiven` to be used in boolean contexts (like
        `if` statements) to signify the absence of a value. This is especially useful
        for checking if an argument was provided or not in a function.
        """
        return False

    @override
    def __repr__(self) -> Literal["NOT_GIVEN"]:
        return "NOT_GIVEN"

    def __setattr__(self, key: str, value: Any) -> None:
        raise AttributeError(f"{self.__class__.__name__} instances are immutable")

    def __delattr__(self, key: str) -> None:
        raise AttributeError(f"{self.__class__.__name__} instances are immutable")


NOT_GIVEN = _NotGiven()


class _Missing:
    """
    -   **Primary Use:** `MISSING` is more common in data structures,
        configurations, or APIs where you need to signify that a value hasn't been
        set or provided, and it's expected to be present or filled in later.
    -   **Semantics:** It indicates the absence of a value in a more passive sense,
        as in "not yet provided" or "awaiting assignment."
    -   **Example:** In a configuration object, `None` might be used to disable an
        option, whereas `MISSING` would indicate that the user has not yet made a
        decision about that option.
    """

    _instance: _Missing | None = None

    def __new__(cls: Type[_Missing]) -> _Missing:  # noqa: PYI034
        if cls._instance is None:
            cls._instance = super(_Missing, cls).__new__(cls)  # noqa: UP008
        return cls._instance

    def __bool__(self) -> Literal[False]:
        return False

    def __repr__(self) -> Literal["MISSING"]:
        return "MISSING"

    def __setattr__(self, key: str, value: Any) -> None:
        raise AttributeError(f"{self.__class__.__name__} instances are immutable")

    def __delattr__(self, key: str) -> None:
        raise AttributeError(f"{self.__class__.__name__} instances are immutable")


MISSING = _Missing()


class _Omit:
    """
    In certain situations you need to be able to represent a case where a default
    value has to be explicitly removed and `None` is not an appropriate substitute,
    for example:

    ```python
    # as the default `Content-Type` header is `application/json` that will be sent
    client.post('/upload/files', files={'file': b'my raw file content'})

    # you can't explicitly override the header as it has to be dynamically generated
    # to look something like:
    # 'multipart/form-data; boundary=0d8382fcf5f8c3be01ca2e11002d2983'
    client.post(..., headers={'Content-Type': 'multipart/form-data'})

    # instead you can remove the default `application/json` header by passing Omit
    client.post(..., headers={'Content-Type': Omit()})
    ```
    """

    _instance: _Omit | None = None

    def __new__(cls: Type[_Omit]) -> _Omit:  # noqa: PYI034
        if cls._instance is None:
            cls._instance = super(_Omit, cls).__new__(cls)  # noqa: UP008
        return cls._instance

    def __bool__(self) -> Literal[False]:
        return False

    def __repr__(self) -> Literal["OMIT"]:
        return "OMIT"

    def __setattr__(self, key: str, value: Any) -> None:
        raise AttributeError(f"{self.__class__.__name__} instances are immutable")

    def __delattr__(self, key: str) -> None:
        raise AttributeError(f"{self.__class__.__name__} instances are immutable")


OMIT = _Omit()


class Singleton(type, Generic[_T]):
    """Singleton metaclass for creating singleton classes.

    References
    ----------
    [1] https://stackoverflow.com/questions/6760685/what-is-the-best-way-of-implementing-singleton-in-python
    [2] https://stackoverflow.com/questions/75307905/python-typing-for-a-metaclass-singleton
    """

    _instances: Dict[Singleton[_T], _T] = {}
    _lock: threading.Lock = threading.Lock()  # Create a lock object

    def __call__(cls: Singleton[_T], *args: Any, **kwargs: Any) -> _T:
        # Lock the block of code where the instance is checked and created
        with cls._lock:
            if cls not in cls._instances:
                cls._instances[cls] = super().__call__(*args, **kwargs)
            return cls._instances[cls]



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, Type, Union

from pydantic import BaseModel, Field, field_validator

from omnivault._types._sentinel import MISSING


class LoggerConfig(BaseModel):
    """The data config."""

    log_file: Union[str, None] = None
    module_name: Union[str, None] = None
    propagate: bool = False
    log_root_dir: Union[str, None] = None
    rich_handler_config: Dict[str, Any] = Field(
        default_factory=lambda: {
            "level": "INFO",  # logging.INFO,
            "console": MISSING,
            "show_level": True,
            "show_path": True,
            "show_time": True,
            "rich_tracebacks": True,
            "markup": True,
            "log_time_format": "[%Y-%m-%d %H:%M:%S]",
        }
    )

    @field_validator("log_root_dir")
    @classmethod
    def check_log_root_dir(cls: Type[LoggerConfig], v: str) -> str:
        if v is not None:
            path = Path(v)
            if not path.exists():
                path.mkdir(parents=True, exist_ok=True)
        return v



"""
This package, omnivault.dsa.searching_algorithms, implements various searching
algorithms following the Strategy Design Pattern. It includes a context class,
multiple search strategies, and a client interface for executing searches.

Modules:

- base.py: Defines base classes and common functionalities for searching
  algorithms.
- context.py: Contains the Context class that interacts with different search
  strategies.
- strategies.py: Implements various search strategies as per the Strategy Design
  Pattern.
- client.py: Demonstrates the usage of search strategies within a client
  application.
"""



r"""
Scaled Dot-Product Attention Module
===================================

This module defines the Scaled Dot-Product Attention mechanism, which is a key
component in transformer architectures. It follows the equations:

.. math::
    \text{Attention}(\mathbf{Q}, \mathbf{K}, \mathbf{V}) = \text{softmax} \left( \frac{\mathbf{Q}\mathbf{K}^T}{\sqrt{d_k}} \right) \mathbf{V}

The attention mechanism is computed as follows:

1. Find the embedding dimension $D$ or $d_q$ from the query feature vector. Note
   the query vector has the below representation:

   .. math::
       \mathbf{Q}   = \mathbf{Z} @ \mathbf{W}^{\mathbf{Q}} \in \mathbb{R}^{T \times D}
       \mathbf{Q}_h = \mathbf{Q} @ \mathbf{W}_{h}^{\mathbf{Q}} \in \mathbb{R}^{T \times d_q}
       d_q = \frac{D}{H}

    However, we note :math:\mathbf{Q}_h is more of a symbolic understanding
    than a real implementation.

2. Compute the dot product of the query feature vector :math:`\mathbf{Q}` with
   the transpose of the key feature vector :math:`\mathbf{K}`.
   Note since in practice the query, key and value are batched, we would have
   :math:`\mathbf{K} \in \mathbb{R}^{B \times T \times D}` and so we operate the
   transpose on the last two dimensions, specified by `dim0` and `dim1`.
   key.transpose(dim0=-2, dim1=-1) means let the second last dimension be the
   last dimension and let the last dimension be the second last dimension.

   This is the attention scores. Initially, the raw attention scores are scaled
   by the inverse square root of the dimensionality of the key vectors to
   stabilize the gradients during training.

3. Apply mask to the scores if mask is not None. Softmax of -INF is zero,
   effectively zeroing out masked logits. If False in mask is we ignore, then we
   should fill attention scores with -INF. masked_fill fills tensor with value
   -INF whenever mask's value is True. So we want the converse, so mask==0 means
   whenever mask has False, then mask==0 evaluates to True, and we fill that.

   Now attention scores are filled with -INF whenever mask is False.

4. Context vector formation: The final context vectors are calculated as a
weighted sum of the value vectors, with the weights provided by the
attention weights. This step effectively combines the value vectors based on
the computed attention, focusing the model's attention on certain elements.

    .. math::
        \text{context_vector} = \text{attention_weights} \cdot V

5. Output generation: The computed context vectors and the attention weights
are then output from the attention mechanism. The context vectors serve as a
synthesis of the input sequence as seen by the model, incorporating the most
relevant pieces of information according to the attention weights.

Self-Attention versus Cross-Attention
=====================================

In the context of attention mechanisms, especially in sequence-to-sequence
models like transformers, the terms "query", "key", and "value" can come from
different parts of the model and have different sequence lengths:

- The "query" usually comes from the decoder side of a transformer model. The
  sequence length of the query is often referred to as $T$ (target sequence
  length) because it relates to the target sequence that the model is trying to
  generate.
- The "key" and "value" usually come from the encoder side. The sequence length
  of the key and value is often referred to as $S$ (source sequence length)
  because they relate to the source sequence that is being encoded.

This distinction becomes particularly important in cross-attention modules where
the decoder (target) needs to attend to the output of the encoder (source). In
self-attention modules within either the encoder or decoder, the sequence
lengths of query, key, and value would be the same, and this distinction
wouldn't be necessary.

Multi-Head Attention Module
===========================

This module defines the Multi-Head Attention mechanism, which is a key
component in transformer architectures. It follows the equations:

.. math::
    \text{MultiHead}(\mathbf{Q}, \mathbf{K}, \mathbf{V}) = \text{Concat}(\text{head}_1, ..., \text{head}_H)W^O

where each head is computed as:

.. math::
    \text{head}_h = \text{Attention}(QW^Q_h, KW^K_h, VW^V_h) where h \in [1, H]

Here, \mathbf{Q}, \mathbf{K}, \mathbf{V} are the query, key, and value vectors. W^Q_h, W^K_h, W^V_h are
parameter matrices. H is the number of heads.

Notations
=========

- For a decoder-only model (like GPT), where self-attention is used, the
  sequence lengths for query, key, and value are the same. It's common to use a
  single term for the sequence length, such as $L$ or $T$, because there is no
  need to differentiate between different sequence lengths within the attention
  mechanism.

- For a full encoder-decoder model (like the Transformer), it is beneficial to
  differentiate the sequence lengths. Here, $T$ typically denotes the sequence
  length of the query vectors that come from the decoder, and $S$ denotes the
  sequence length of the key and value vectors that come from the encoder. This
  distinction is important during cross-attention, where the decoder attends to
  the output of the encoder, and thus the lengths can be different.
"""
from __future__ import annotations

from typing import Tuple

import torch
from torch import nn

from omnivault.transformer.modules.attention.base import Attention

__all__ = ["MultiHeadedAttention", "ScaledDotProductAttention"]


class ScaledDotProductAttention(Attention):
    r"""Implements scaled dot-product attention mechanism.

    This class is a derived instance of the `Attention` class that computes the
    scaled dot-product attention, defined by the following operation:

    .. math::
        \text{Attention}(\mathbf{Q}, \mathbf{K}, \mathbf{V}) = \text{softmax} \left( \frac{QK^T}{\sqrt{d_k}} \right) \mathbf{V}

    where:

    -   Q is the query matrix
    -   K is the key matrix
    -   V is the value matrix
    -   d_k is the dimension of the keys

    Note
    ----
    The attention mechanism can be applied in two different contexts: self-attention
    and cross-attention. Self-attention allows the model to integrate information
    from the entire sequence, while cross-attention allows the model to focus on
    information from a different sequence (e.g., encoder outputs).

    Consequently, we have the following implicit assumptions:

    -   For self-attention, the sequence lengths of `query`, `key`, and `value` are
        the same. This is because the query, key, and value are all derived from the
        same sequence `z` (e.g., the output of the token embedding and positional
        encoding layers).

    1. **Self-Attention** (also known as intra-attention) is used within a single
    sequence to relate different positions of this sequence. In self-attention,
    the queries (Q), keys (K), and values (V) all come from the same previous
    layer output. This is common in both the encoder and the decoder layers of a
    Transformer. For example, in a standalone GPT-Decoder variant, the initial
    queries, keys, and values all originates from the same input sequence, which
    becomes the `z` in the token embedding and positional encoding layers.

    2. **Cross-Attention** is used in the decoder to attend to the output of the
    encoder. In cross-attention, the queries (Q) come from the previous layer of
    the decoder, but the keys (K) and values (V) come from the output of the
    encoder. This allows each position in the decoder to attend to all positions
    in the input sequence.
    """

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        mask: torch.BoolTensor | None = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Perform the forward pass for scaled dot-product attention.

        This function applies the attention mechanism on the input tensors `query`,
        `key`, and `value`. It's worth noting that for cross-attention, the sequence
        lengths of `query` and `key`/`value` may differ. This is because `query` is
        usually projected from the decoder's states, while `key` and `value` are from
        the encoder's states.

        Notations
        ---------
        - B   : Batch size
        - D   : Embedding dimension
        - H   : Number of heads
        - d_k : Dimension of the keys    = D // H
        - d_q : Dimension of the queries = D // H
        - d_v : Dimension of the values  = D // H
        - T   : Sequence length for `query`
        - S   : Sequence length for `key` and `value`
        - L   : Sequence length for `query`, `key` and `value` generic.

        Note
        ----
        -   We use `L` in our notes instead of `T` and `S` since we assume all query,
            key and value are of same length as we are dealing with self-attention in
            GPT decoder.

        -   We often denote the dimension of the keys and queries as `d_k` instead of
            `d_k` and `d_q` respectively because both must have the same dimensionality
            for them to be multiplied together.

        Parameters
        ----------
        query:  A tensor of query vectors representing the set of elements each sequence
                is seeking to attend to. It contains a batch of sequences, each with a set of
                vectors across multiple attention heads.
                    type :  torch.Tensor
                    shape: `(B, H, T, d_q)` where `d_q = D // H`
                    shape: `(B, H, L, d_q)` if in pure self-attention (GPT)
        key  :  A tensor of key vectors that are paired with values to form a mapping. The
                dot product of a query with these keys determines the attention weight for the
                corresponding values.
                    type :  torch.Tensor
                    shape: `(B, H, S, d_k)` where `d_k = D // H`
                    shape: `(B, H, L, d_k)` if in pure self-attention (GPT)
        value: A tensor of value vectors that are aggregated based on the attention
               weights to form the output of the attention mechanism.
                    type :  torch.Tensor
                    shape: `(B, H, S, d_v)` where `d_v = D // H`
                    shape: `(B, H, L, d_v)` if in pure self-attention (GPT)
        mask : An optional boolean mask tensor that can be used to mask out certain positions
               from the attention mechanism. For self-attention, the mask shape is typically
               `(B, T, T)` or `(B, L, L). For cross-attention, the mask typically has a shape of
               `(B, T, S)` allowing different target positions to attend to different source positions.
               Here, `T` is the sequence length of the queries (note `T` is the same for
               self-attention), `S` is the sequence lengths of the keys and values
               which could be equal to `T` in self-attention or vary in cross-attention, and
               `S` is the sequence length of the source (encoder) when using cross-attention.

               However, to cater to the head dimension `H`, right after the `B` dimension, we
               will need to add (unsqueeze) an dimension to have `(B, 1, T, T)` for
               self-attention and `(B, 1, T, S)` for cross-attention.

        Returns
        -------
        context_vector, attention_weights: Tuple[torch.Tensor, torch.Tensor]
            The context vectors and the attention weights. The context vectors are the weighted sum
            of the `value` vectors, representing the information to be attended to.
            The attention weights represent the attention probabilities.

            - Context Vectors shape:   `(B, T, d_v)`
            - Attention Weights shape: `(B, T, S)` or `(B, H, T, S)` or `(B, H, L, L)` if in pure self-attention (GPT)
        """
        # fmt: off
        d_q               = query.size(dim=-1)

        attention_scores  = torch.matmul(query, key.transpose(dim0=-2, dim1=-1)) / torch.sqrt(torch.tensor(d_q).float())        # [B, H, T, d_q] @ [B, H, d_q, T] = [B, H, T, T]
        attention_scores  = attention_scores.masked_fill(mask == 0, float("-inf")) if mask is not None else attention_scores    # [B, H, T, T]

        attention_weights = attention_scores.softmax(dim=-1)        # [B, H, T, T]
        attention_weights = self.dropout(attention_weights)         # [B, H, T, T]

        context_vector    = torch.matmul(attention_weights, value)  # [B, H, T, T] @ [B, H, T, d_v] = [B, H, T, d_v]
        # fmt: on
        return context_vector, attention_weights


class MultiHeadedAttention(nn.Module):
    __slots__ = [
        "d_model",
        "d_k",
        "d_q",
        "d_v",
        "H",
        "W_Q",
        "W_K",
        "W_V",
        "W_O",
        "attention",
        "dropout",
        "context_vector",
        "attention_weights",
    ]

    def __init__(
        self,
        attention: Attention,
        H: int,
        d_model: int,
        dropout: float = 0.1,
        bias: bool = False,
    ) -> None:
        super().__init__()
        assert d_model % H == 0, "The number of heads must divide the embedding dimension."

        # fmt: off
        self.d_model   = d_model       # D
        self.d_k       = d_model // H  # stay true to notations
        self.d_q       = d_model // H
        self.d_v       = d_model // H

        self.H         = H             # number of heads

        # shadow my notations, actually they are of shape D x D.
        self.W_Q       = nn.Linear(self.d_model, self.d_q * self.H, bias=bias)  # D x D
        self.W_K       = nn.Linear(self.d_model, self.d_k * self.H, bias=bias)
        self.W_V       = nn.Linear(self.d_model, self.d_v * self.H, bias=bias)
        self.W_O       = nn.Linear(self.d_model, self.d_model, bias=bias)

        self.attention = attention
        self.dropout   = nn.Dropout(p=dropout, inplace=False)

        self.context_vector: torch.Tensor
        self.attention_weights: torch.Tensor

        # self._init_weights()
        # fmt: on

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        mask: torch.BoolTensor | None = None,
    ) -> torch.Tensor:
        """
        Notations
        ---------
        B:      Batch size
        S or L: Source sequence length
        T or L: Target sequence length
        D:      Embedding dimension
        H:      Number of heads

        Parameters
        ----------
        query:  Although named as query, it is the embeddings `z` from the token_embedding + positional_embedding layer.
                type:  torch.Tensor
                shape: (B, S or T, D)
        key:    Although named as key, it is the embeddings `z` from the token_embedding + positional_embedding layer.
                type:  torch.Tensor
                shape: (B, S or T, D)
        value:  Although named as value, it is the embeddings `z` from the token_embedding + positional_embedding layer.
                type:  torch.Tensor
                shape: (B, S or T, D)
        mask:   Mask to be applied to the attention scores.
                type:  torch.BoolTensor
                shape: (B, 1, S or T, S or T)

        Returns
        -------
        O:  The output of the multi-headed attention mechanism.
            type:  torch.Tensor
            shape: (B, S or T, D)

        Variables
        ---------
        W_Q.weight (D, D)
        W_K.weight (D, D)
        W_V.weight (D, D)
        W_O.weight (D, D)
        """
        # fmt: off
        if mask is not None:
            assert mask.ndim     == 4, f"Mask should have 4 dimensions but got {mask.ndim}."
            assert mask.shape[0] == query.shape[0], ("Batch size of mask and query must match.")
            assert mask.shape[1] == 1, ("Mask should have shape (batch_size, 1, seq_len, seq_len).")
            assert mask.shape[2] == mask.shape[3] == query.shape[1], ("Mask should have shape (batch_size, 1, seq_len, seq_len).")


        Q = self.W_Q(query).contiguous() # Z @ W_Q -> BxTxD @ DxD = BxTxD
        K = self.W_K(key).contiguous()   # Z @ W_K
        V = self.W_V(value).contiguous() # Z @ W_V

        Q = self.transpose_qkv(Q)        # splitting happens -> [B, H, L, D]
        K = self.transpose_qkv(K)
        V = self.transpose_qkv(V)

        # Attention
        self.context_vector, self.attention_weights = self.attention(Q, K, V, mask)
        context_vector_concat                       = self.reverse_transpose_qkv(self.context_vector)
        # fmt: on

        # mypy complains because it infers `O` as `Any` but it is actually a tensor.
        # You can either cast it to tensor or use `self.W_O.forward(context_vector_concat)`.
        O = self.W_O(context_vector_concat)  # context_vector_concat @ W_O -> LxD @ DxD = LxD
        return O  # type: ignore[no-any-return]

    def _init_weights(self) -> None:
        """See PyTorch's MultiHeadAttention code for reference."""
        # we assume _qkv_same_embed_dim is True
        nn.init.xavier_uniform_(self.W_Q.weight)
        nn.init.xavier_uniform_(self.W_K.weight)
        nn.init.xavier_uniform_(self.W_V.weight)
        nn.init.xavier_uniform_(self.W_O.weight)

    def transpose_qkv(self, q_or_k_or_v: torch.Tensor) -> torch.Tensor:
        """Transposition for parallel computation of multiple attention heads.
        Why does transpose allow parallel computation? So originally the shape of
        the query, key, and value is (B, L, D), and we want to split the D into H
        heads to become (B, L, H, D / H). But this is not the shape we want (could
        be due to efficiency reasons), so we transpose the shape to (B, H, L, D / H)
        so all heads can be computed in parallel (efficiently).

        Parameters
        ----------
        q_or_k_or_v: The query, key, or value tensor.
            type:  torch.Tensor
            shape: (B, L, D)

        Returns
        -------
        q_or_k_or_v: The transposed query, key, or value tensor.
            type:  torch.Tensor
            shape: (B, H, L, D / H)
        """
        # fmt: off
        # 1. q_or_k_or_v is shape (B, L, D)
        # 2. aim to make it of shape (B, L, H, D / H = d_qkv)
        batch_size, seq_len, _ = q_or_k_or_v.shape
        q_or_k_or_v            = q_or_k_or_v.view(batch_size, seq_len, self.H, self.d_model // self.H)

        # 3. switch H from 3rd to 2nd dimension, or in python swap 2nd to 1st dimension and 1st to 2nd dimension
        #    shape (B, H, L, D / H = d_qkv)
        q_or_k_or_v            = q_or_k_or_v.permute(0, 2, 1, 3)
        # fmt: on
        return q_or_k_or_v

    def reverse_transpose_qkv(self, q_or_k_or_v: torch.Tensor) -> torch.Tensor:
        """Reverse the transposition operation for concatenating multiple attention heads.

        Parameters
        ----------
        q_or_k_or_v: The query, key, or value tensor.
            type:  torch.Tensor
            shape: (B, H, L, D / H)

        Returns
        -------
        q_or_k_or_v: The transposed query, key, or value tensor.
            type:  torch.Tensor
            shape: (B, L, D)
        """
        # fmt: off
        # 1. q_or_k_or_v is shape (B, H, L, D / H = d_qkv)
        # 2. aim to make it of shape (B, L, H, D / H = d_qkv)
        q_or_k_or_v = q_or_k_or_v.permute(0, 2, 1, 3)

        # 3. Merge H and d_qkv into D
        batch_size, seq_len, _, _ = q_or_k_or_v.shape
        q_or_k_or_v = q_or_k_or_v.contiguous().view(batch_size, seq_len, self.d_model)
        # fmt: on
        return q_or_k_or_v



"""
This module provides utilities for creating quiver plots using Matplotlib.

A quiver plot is a type of plot that shows vector fields, which are useful for
visualizing velocity fields, gradients, or any other kind of directional data.
The plot consists of arrows (vectors) placed at various points in a 2D space.

Parameters for creating a quiver plot:
- X, Y: These arrays define the starting positions (origins) of the vectors.
  Each arrow in the plot originates from the coordinate (X[i], Y[i]). This
  represents the 'origin of the basis' for each vector in the field. X and Y
  should have the same shape and can be generated using meshgrid for regular
  grids or specified manually for irregular grids.
- U, V: These arrays represent the vector components at each point defined by X
  and Y. U[i], V[i] are the horizontal (x-direction) and vertical (y-direction)
  displacements from the origin (X[i], Y[i]). These determine the direction and
  magnitude (length) of each vector. The length of the vector is proportional
  to the magnitude of the displacement, calculated as sqrt(U[i]**2 + V[i]**2).

NOTE
----
All `quiver` calls only assumes that its takes in float values instead of
`1D` or `2D` array-like objects. This means that the `X`, `Y`, `U`, and `V`
are floats unpacked from our `Vector` objects. We then use a for loop to
iterate over the vectors and plot them one by one. Maybe consider FIXME
this to support vectorized plotting.
"""
from __future__ import annotations

from typing import Any, Dict, List, Optional, Sequence

import matplotlib.pyplot as plt

from omnivault._types._generic import Vec
from omnivault.linear_algebra.base import VectorPlotter
from omnivault.linear_algebra.vector import Vector2D, Vector3D


def add_vectors_to_plotter(plotter: VectorPlotter[Vec], vectors: Sequence[Vec]) -> None:
    """Add vectors to a plotter.

    Type
    ----
    The reason we use `Vec` instead of `Vector` is because `Vec` is a type
    variable bounded to `Vector` and thus can represent `Vector` and any
    of its subclasses. This means `Vec` is an upper bound for `Vector` and
    its subclasses.
    """
    for vector in vectors:
        plotter.add_vector(vector)


def add_text_annotations(
    plotter: VectorPlotter[Vec],
    vectors: Sequence[Vec],
    include_endpoint_label: bool = True,
    include_vector_label: bool = True,
    endpoint_kwargs: Optional[Dict[str, Any]] = None,
    vector_kwargs: Optional[Dict[str, Any]] = None,
) -> None:
    endpoint_kwargs = endpoint_kwargs or {"fontsize": 12}
    vector_kwargs = vector_kwargs or {"fontsize": 12}

    for vector in vectors:
        dim = len(vector)  # Determine if the vector is 2D or 3D

        # endpoint = tuple(map(sum, zip(vector.origin, vector.direction)))
        if include_endpoint_label:
            # Label with endpoint coordinates
            x_end = vector.origin[0] + vector.direction[0]
            y_end = vector.origin[1] + vector.direction[1]
            if dim == 3:
                z_end = vector.origin[2] + vector.direction[2]
                label = f"({x_end}, {y_end}, {z_end})"
                plotter.add_text(x=x_end, y=y_end, z=z_end, text=label, **endpoint_kwargs)
            else:
                label = f"({x_end}, {y_end})"
                plotter.add_text(x=x_end, y=y_end, text=label, **endpoint_kwargs)

        if include_vector_label and vector.label:
            midpoint = tuple(origin + direction / 2 for origin, direction in zip(vector.origin, vector.direction))
            # Label with vector label
            if dim == 3:
                plotter.add_text(
                    x=midpoint[0],
                    y=midpoint[1],
                    z=midpoint[2],
                    text=vector.label,
                    **vector_kwargs,
                )
            else:  # dim == 2
                plotter.add_text(x=midpoint[0], y=midpoint[1], text=vector.label, **vector_kwargs)


class VectorPlotter2D(VectorPlotter[Vector2D]):
    def __init__(
        self,
        fig: Optional[plt.Figure] = None,
        ax: Optional[plt.Axes] = None,
        ax_kwargs: Optional[Dict[str, Dict[str, Any]]] = None,
        quiver_kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(fig, ax, ax_kwargs)

        self.quiver_kwargs = quiver_kwargs or {
            "angles": "xy",
            "scale_units": "xy",
            "scale": 1,
            "alpha": 0.6,
        }

        self.vectors: List[Vector2D] = []
        self.colors: List[str] = []

        self._apply_ax_customizations()  # may not need since in FigureManager

    def add_text(
        self,
        x: float,
        y: float,
        text: str,
        z: Optional[float] = None,  # noqa: ARG002
        fontsize: int = 16,
        **kwargs: Any,
    ) -> None:
        self.ax.text(x, y, text, fontsize=fontsize, **kwargs)

    def annotate(
        self,
        x: float,
        y: float,
        text: str,
        arrow_props: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        self.ax.annotate(
            text,
            xy=(x, y),
            xytext=(x, y),
            arrowprops=arrow_props,
            fontsize=16,
            **kwargs,  # type: ignore[arg-type]
        )

    def add_vector(self, vector: Vector2D) -> None:
        self.vectors.append(vector)

    def plot(self, grid: bool = True, show_ticks: bool = False) -> None:
        """Currently only works in notebooks, if in script, please add
        `plt.show()` after calling this method."""
        for vector in self.vectors:
            # fmt: off
            X, Y = vector.origin
            U, V = vector.direction
            # fmt: on
            self.ax.quiver(X, Y, U, V, color=vector.color, **self.quiver_kwargs)

        if grid:
            self.ax.grid()

        if not show_ticks:
            self.ax.tick_params(axis="both", which="both", length=0)


class VectorPlotter3D(VectorPlotter[Vector3D]):
    from mpl_toolkits.mplot3d import Axes3D  # type: ignore[import-untyped]

    def __init__(
        self,
        fig: Optional[plt.Figure] = None,
        ax: Optional[plt.Axes] = None,
        ax_kwargs: Optional[Dict[str, Dict[str, Any]]] = None,
        quiver_kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(fig, ax, ax_kwargs)

        self.quiver_kwargs = quiver_kwargs or {
            "length": 1,
            "normalize": False,
            "alpha": 0.6,
            "arrow_length_ratio": 0.18,
            "pivot": "tail",
            "linestyles": "solid",
            "linewidths": 3,
        }

        self.vectors: List[Vector3D] = []
        self.colors: List[str] = []

        self._apply_ax_customizations()  # may not need since in FigureManager

    def add_text(
        self,
        x: float,
        y: float,
        text: str,
        z: Optional[float] = None,
        fontsize: int = 16,
        **kwargs: Any,
    ) -> None:
        self.ax.text(x, y, z, text, fontsize=fontsize, **kwargs)  # type: ignore[arg-type]

    def annotate(
        self,
        x: float,
        y: float,
        z: float,
        text: str,
        arrow_props: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        self.ax.annotate(
            text,
            xy=(x, y, z),  # type: ignore[arg-type]
            xytext=(x, y, z),  # type: ignore[arg-type]
            arrowprops=arrow_props,
            fontsize=16,
            **kwargs,  # type: ignore[arg-type]
        )

    def add_vector(self, vector: Vector3D) -> None:
        self.vectors.append(vector)

    def plot(self, grid: bool = True, show_ticks: bool = False) -> None:
        for vector in self.vectors:
            # fmt: off
            X, Y, Z = vector.origin
            U, V, W = vector.direction
            # fmt: on
            self.ax.quiver(X, Y, Z, U, V, W, color=vector.color, **self.quiver_kwargs)

        if grid:
            self.ax.grid()

        if not show_ticks:
            self.ax.tick_params(axis="both", which="both", length=0)



import json
import pickle
from typing import Any


def dump_json(data: Any, filepath: str, **kwargs: Any) -> None:
    """
    Dumps data to a JSON file.

    Parameters
    ----------
    data : Any
        The Python data to dump to the JSON file.
    filepath : str
        The path to the JSON file to write to.
    kwargs: Any
        Additional keyword arguments for `json.dump()`.
    """
    with open(filepath, "w") as file:
        json.dump(data, file, **kwargs)


def load_json(filepath: str, **kwargs: Any) -> Any:
    """
    Loads data from a JSON file.

    Parameters
    ----------
    filepath : str
        The path to the JSON file to read from.
    kwargs: Any
        Additional keyword arguments for `json.load()`.

    Returns
    -------
    Any
        The Python data loaded from the JSON file.
    """
    with open(filepath, "r") as file:
        return json.load(file, **kwargs)


def pickle_dump(obj: Any, filepath: str) -> None:
    """
    Serializes an object to a binary file using pickle.

    Parameters
    ----------
    obj : Any
        The object to serialize.
    filepath : str
        The path to the file where the object will be stored.
    """
    with open(filepath, "wb") as file:
        pickle.dump(obj, file)


def pickle_load(filepath: str) -> Any:
    """
    Deserializes an object from a binary file using pickle.

    Parameters
    ----------
    filepath : str
        The path to the file from which the object will be loaded.

    Returns
    -------
    Any
        The deserialized Python object.
    """
    with open(filepath, "rb") as file:
        return pickle.load(file)



from __future__ import annotations

import gc
from typing import List

import torch


def purge_global_scope(variable_name_or_names: str | List[str]) -> None:
    """
    Deletes the provided objects and performs cleanup.

    Parameters
    ----------
    objects: List[Any]
        The list of objects to be deleted.

    Notes
    -----
    The below does not work because we are passing in references of the
    variables and are not really deleted.

    ```python
    def purge_global_scope(object_or_objects: Any | List[Any]) -> None:
        if isinstance(object_or_objects, list):
            for obj in object_or_objects:
                del obj
        else:
            del object_or_objects

        gc.collect()
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
    ```
    """
    if isinstance(variable_name_or_names, str):
        variable_name_or_names = [variable_name_or_names]

    global_vars = globals()
    for name in variable_name_or_names:
        if name in global_vars:
            del global_vars[name]

    gc.collect()
    if torch.cuda.is_available():
        torch.cuda.empty_cache()



import importlib
from types import ModuleType
from typing import Dict, Generic, Tuple, Type

from pydantic import BaseModel, Field

from omnivault._types._generic import DynamicClass, T, V


class DynamicClassFactory(BaseModel, Generic[DynamicClass]):
    r"""
    A factory class for dynamic instantiation of classes based on a configuration.

    Using a `TypeVar` for the generic type `DynamicClass` is better than using `Any`
    since we really have no way to know the return type of `create_instance()`.

    This class serves as a base for configuration management, particularly useful
    in scenarios involving complex systems or large libraries where classes or
    functions come with numerous arguments. It centralizes the management of
    these configurations, simplifying instantiation and enhancing flexibility and
    maintainability of the code.

    Attributes
    ----------
    name : str
        Fully qualified class name to be instantiated. This should include
        both the module and the class name (e.g., 'mymodule.MyClass'). The factory
        dynamically imports the module, accesses the class, and creates an
        instance using other configuration parameters provided in the subclass.

    Methods
    -------
    pop_name_and_return_remaining_kwargs() -> KwargsDict:
        Extracts the configuration parameters excluding 'name' attribute.

    create_instance() -> Any:
        Dynamically creates an instance of the class specified in 'name' attribute
        using the provided configuration parameters.

    Note
    ----
    - This class is particularly beneficial for larger, complex systems where
      configurations need to be changed frequently or shared across components.
    - It is designed to be extended by specific configuration classes tailored to
      particular use cases.

    Examples
    --------
    >>> class TextSplitterConfig(DynamicClassFactory):
    ...     name: str = "langchain.text_splitter.RecursiveCharacterTextSplitter"
    ...     separators: List[str] = Field(default_factory=lambda: ["\n\n", "\n", " ", ""])
    ...     chunk_size: int = 300
    ...     chunk_overlap: int = 50
    ...     length_function: Callable = Field(default_factory=lambda: len)
    >>> splitter_config = TextSplitterConfig()
    >>> text_splitter = splitter_config.create_instance()
    """

    name: str = Field(
        ...,
        description=(
            "The 'name' field should contain the fully qualified class name "
            "that this factory will instantiate. This includes both the module "
            "and the class name. For example, 'mymodule.MyClass'. The factory "
            "will dynamically import the specified module, access the class, "
            "and create an instance of it using the other configuration "
            "parameters provided in the subclass."
        ),
    )

    def pop_name_and_return_remaining_kwargs(self) -> Dict[str, V]:
        return {key: value for key, value in self.model_dump(mode="python").items() if key != "name"}

    def split_class_name(self) -> Tuple[str, str]:
        """Split the full class name into module and class names."""
        # pylint: disable=no-member
        if not self.name or self.name.startswith(".") or self.name.endswith("."):
            raise ValueError(f"Invalid class name format: {self.name}")

        try:
            module_name, class_name = self.name.rsplit(".", 1)
            return module_name, class_name
        except ValueError as err:
            raise ValueError(f"Invalid class name format: {self.name}") from err

    def import_module(self, module_name: str) -> ModuleType:
        """Import and return the module based on the module name."""
        try:
            return importlib.import_module(module_name)
        except ImportError as err:
            raise ImportError(f"Module {module_name} cannot be imported") from err

    def get_class_from_module(self, module: ModuleType, class_name: str) -> Type[DynamicClass]:
        """Retrieve the class from the imported module."""
        try:
            return getattr(module, class_name)  # type: ignore[no-any-return]
        except AttributeError as err:
            raise AttributeError(f"Class {class_name} not found in module {module.__name__}") from err

    def create_instance(self, *args: T, **kwargs: T) -> DynamicClass:
        """
        Note
        ----
        1.  self.name is an instance of FieldInfo but at run time it should be
            str, so I disable pylint complaints.
        2.  args and kwargs are passed to the class constructor to handle cases
            such as PyTorch Optimizer that requires parameters to be passed in
            the constructor.
        """
        # fmt: off
        module_name, class_name = self.split_class_name()
        module                  = self.import_module(module_name)
        class_                  = self.get_class_from_module(module, class_name)
        # fmt: on
        return class_(*args, **self.pop_name_and_return_remaining_kwargs(), **kwargs)



from __future__ import annotations

from abc import ABC, abstractmethod

import torch


class Activation(ABC):
    """
    Base class for activation functions.
    """

    def __init__(self, dim: int | None = None) -> None:
        """
        Initialize the activation function.
        """
        self.dim = dim

    @abstractmethod
    def __call__(self, z: torch.Tensor) -> torch.Tensor:
        """
        Compute the output of the activation function.
        """
        raise NotImplementedError

    def gradient(self, z: torch.Tensor) -> torch.Tensor:
        """
        Compute the gradient of the activation function.
        """
        raise NotImplementedError



from typing import Any


class ImmutableProxy:
    """Immutable proxy object for wrapping mutable objects.

    Example
    -------
    >>> from omnivault.utils.general import ImmutableProxy
    >>> obj = ImmutableProxy([1, 2, 3])
    >>> obj.append(4)
    Traceback (most recent call last):
        ...
        AttributeError: Attempting to set attribute append with <built-in method append of list object at xxx>. list object is immutable.
    """

    def __init__(self, obj: object) -> None:
        object.__setattr__(self, "_obj", obj)

    def __getattr__(self, item: str) -> Any:
        attr = getattr(self._obj, item)
        if callable(attr):

            def immutable_method(*args: Any, **kwargs: Any) -> None:  # noqa: ARG001
                raise AttributeError(
                    f"Attempting to modify object with method `{item}`. "
                    f"`{self._obj.__class__.__name__}` object is immutable."
                )

            return immutable_method
        return attr

    def __setattr__(self, key: str, value: Any) -> None:
        raise AttributeError(
            f"Attempting to set attribute `{key}` with `{value}`. "
            f"`{self._obj.__class__.__name__}` object is immutable."
        ) from None



from __future__ import annotations

import logging
from pathlib import Path

from rich.logging import RichHandler


def configure_logger(
    rank: int,
    log_dir: str | None = None,
    log_level: int = logging.INFO,
    log_on_master_or_all: bool = True,
) -> logging.Logger:
    """
    Configure and return a logger for a given process rank.

    This function sets up a logger for the specified process rank, allowing for separate
    logging configurations based on the rank. The logger is configured to write logs to
    a file named `process_{rank}.log` and display logs with severity level INFO and above.

    Parameters
    ----------
    rank : int
        The rank of the process for which the logger is being configured.
    log_dir : str | None, optional
        The directory where the log files will be stored. If None, no log files will be created.
        Default is None.
    log_level : int, optional
        The log level for the logger. Default is logging.INFO.
    log_on_master_or_all : bool, optional
        Determines whether logs should be displayed only on the master rank (rank 0) or on all ranks.
        If True, logs will be displayed only on the master rank. If False, logs will be displayed
        on all ranks. Default is True.


    Returns
    -------
    logging.Logger
        Configured logger for the specified process rank.

    Notes
    -----
    -   The logger is configured to write logs to a file named `process_{rank}.log`
        in the specified `log_dir` directory. This allows for separate log files for
        each process rank.
    -   The reason for writing each rank's logs to a separate file is to avoid the
        non-deterministic ordering of log messages from different ranks in the same
        file.
    -   If `log_on_master_or_all` is True, logs will be displayed only on the master
        rank (rank 0). If False, logs will be displayed on all ranks.
    -   The file handler, if added, will always log messages from all ranks,
        regardless of the `log_on_master_or_all` setting.
    """
    logger = logging.getLogger(f"Process-{rank}")  # set name for each rank
    logger.setLevel(log_level)
    logger.propagate = False  # no propagation to parent

    logger.handlers = []  # clear existing handlers

    formatter = logging.Formatter("%(asctime)s [%(levelname)s]: %(message)s", datefmt="%Y-%m-%d %H:%M:%S")

    # NOTE: # if log_on_master is False, it evaluates first part of conditional
    #         which will log on all ranks, else only on master rank.
    if not log_on_master_or_all or (log_on_master_or_all and rank == 0):
        console_handler = RichHandler()
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)

    if log_dir is not None:
        Path(log_dir).mkdir(parents=True, exist_ok=True)
        log_path = Path(log_dir) / Path(f"process_{rank}.log")
        file_handler = logging.FileHandler(filename=str(log_path))
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    return logger



from __future__ import annotations

import math
from functools import partial

from torch.optim.lr_scheduler import LambdaLR
from torch.optim.optimizer import Optimizer

__all__ = [
    "_get_cosine_schedule_with_warmup_lr_lambda",
    "get_cosine_annealing_with_warmup",
    "_cosine_schedule_with_warmup_and_post_annealing_lr_lambda",
]


def _get_cosine_schedule_with_warmup_lr_lambda(
    current_step: int, *, num_warmup_steps: int, num_training_steps: int, alpha_f: float
) -> float:
    """
    Helper function for calculating the learning rate using cosine annealing
    with warmup.

    Parameters
    ----------
    current_step: int
        The current step in the training process.
    num_warmup_steps: int
        The number of steps for the warmup phase.
    num_training_steps: int
        The total number of training steps.
    alpha_f: float
        The minimum learning rate at the end of the schedule.

    Returns
    -------
    float
        The calculated learning rate.
    """

    if current_step < num_warmup_steps:
        alpha = current_step / max(1, num_warmup_steps)
    else:
        tau_w = (current_step - num_warmup_steps) / num_training_steps
        tau_w = min(1.0, tau_w)
        alpha = alpha_f + (1 - alpha_f) * (1 + math.cos(math.pi * tau_w)) / 2
    return alpha


def get_cosine_annealing_with_warmup(
    optimizer: Optimizer,
    num_warmup_steps: int,
    num_training_steps: int,
    alpha_f: float = 0.1,
    last_epoch: int = -1,
    verbose: bool = False,
) -> LambdaLR:
    """
    Create a schedule with a learning rate that decreases following the values
    of the cosine function between the initial lr set in the optimizer to 0,
    after a warmup period during which it increases linearly between 0 and the
    initial lr set in the optimizer.

    Parameters
    ----------
    optimizer: `~torch.optim.Optimizer`
        The optimizer for which to schedule the learning rate.
    num_warmup_steps: int
        The number of steps for the warmup phase.
    num_training_steps: int
        The total number of training steps.
    alpha_f: float
        The minimum learning rate at the end of the schedule, by default 0.1.
    last_epoch: int
        The index of the last epoch when resuming training, by default -1.
    verbose: bool
        Whether to print the learning rate at every update, by default False.

    Returns
    -------
    `torch.optim.lr_scheduler.LambdaLR`
        The scheduler with the appropriate schedule.

    Examples
    --------
    >>> from torch import nn
    >>> from torch.optim import Adam
    >>> dummy_model = nn.Linear(1, 1)
    >>> optimizer = Adam(dummy_model.parameters(), lr=3e-4)
    >>> scheduler = get_cosine_annealing_with_warmup(optimizer, num_warmup_steps=5, num_training_steps=10, alpha_f=0.5)
    >>> assert isinstance(scheduler, LambdaLR)
    """

    lr_lambda = partial(
        _get_cosine_schedule_with_warmup_lr_lambda,
        num_warmup_steps=num_warmup_steps,
        num_training_steps=num_training_steps,
        alpha_f=alpha_f,
    )
    return LambdaLR(optimizer, lr_lambda, last_epoch, verbose)


def _cosine_schedule_with_warmup_and_post_annealing_lr_lambda(
    iter: int,
    *,
    max_learning_rate: float,
    min_learning_rate: float,
    warmup_iters: int,
    cosine_cycle_iters: int,
) -> float:
    """
    Calculate the learning rate using cosine annealing schedule with warmup.

    .. warning::
        Do not use this in conjunction with `torch.optim.lr_scheduler.LambdaLR`
        because `LambdaLR` expects the `lr_lambda` function to return a
        multiplicative factor relative to the initial learning rate, not the
        absolute learning rate.

    Parameters
    ----------
    iter : int
        The current training iteration.
    max_learning_rate : float
        The maximum learning rate (used at the end of the warmup).
    min_learning_rate : float
        The minimum (final) learning rate after cosine annealing.
    warmup_iters : int
        The number of iterations for the warmup phase.
    cosine_cycle_iters : int
        The total number of iterations for the cosine annealing cycle (including warmup).

    Returns
    -------
    float
        The calculated learning rate for the current training iteration.
    """
    if iter < warmup_iters:  # warmup phase
        return (iter / max(1, warmup_iters)) * max_learning_rate
    elif iter <= cosine_cycle_iters:  # cosine annealing phase
        progress = (iter - warmup_iters) / (cosine_cycle_iters - warmup_iters)
        return min_learning_rate + 0.5 * (max_learning_rate - min_learning_rate) * (1 + math.cos(math.pi * progress))
    else:  # post-annealing phase
        return min_learning_rate



"""
TODO
----
1. Strategy Design Pattern?
2. Integrate with FigureManager.
"""

from typing import Any, List, Optional, Union

import matplotlib.pyplot as plt
import mpl_toolkits
import numpy as np
from matplotlib.axes import Axes
from matplotlib.collections import PathCollection
from matplotlib.container import BarContainer
from matplotlib.lines import Line2D
from matplotlib.quiver import Quiver
from mpl_toolkits.mplot3d import Axes3D
from numpy.typing import NDArray


def plot_line(
    ax: plt.Axes,
    x: NDArray[np.float64],
    y: NDArray[np.float64],
    **kwargs: Any,
) -> List[Line2D]:
    """Plot line."""
    return ax.plot(x, y, **kwargs)


def plot_quiver(
    ax: plt.Axes,  # This can be either a 2D or 3D Axes object
    x: NDArray[np.float64],
    y: NDArray[np.float64],
    u: NDArray[np.float64],
    v: NDArray[np.float64],
    z: Optional[NDArray[np.float64]] = None,  # For 3D
    w: Optional[NDArray[np.float64]] = None,  # For 3D
    **kwargs: Any,
) -> Union[Quiver, mpl_toolkits.mplot3d.art3d.Line3DCollection]:
    """Plot quiver (2D or 3D depending on the ax type)."""
    if isinstance(ax, Axes3D):
        if z is None or w is None:
            raise ValueError("z and w must be provided for 3D quivers")
        return ax.quiver(x, y, z, u, v, w, **kwargs)
    if isinstance(ax, Axes):
        return ax.quiver(x, y, u, v, **kwargs)
    raise ValueError("Invalid Axes type for quiver plot")


def plot_hist(ax: plt.Axes, x: NDArray[np.float64], **kwargs: Any) -> BarContainer:
    """Plot histogram."""
    return ax.hist(x, **kwargs)  # type: ignore


def plot_bar(ax: plt.Axes, x: NDArray[np.float64], y: NDArray[np.float64], **kwargs: Any) -> BarContainer:
    """Plot bar plot."""
    return ax.bar(x, y, **kwargs)


def plot_scatter(ax: plt.Axes, x: NDArray[np.float64], y: NDArray[np.float64], **kwargs: Any) -> PathCollection:
    """Plot scatter plot."""

    return ax.scatter(x, y, **kwargs)



"""This module provides customized random number generator utilities to enhance
reproducibility across different platforms.
"""
import random
from typing import Any, Dict

import numpy as np
import torch

__all__ = ["save_rng_state", "load_and_set_rng_state"]


def save_rng_state(save_dir: str, epoch_index: int) -> None:
    """Save the current state of various random number generators (RNGs) to a file,
    facilitating reproducible experimentation by allowing the exact state of RNGs to
    be restored later. This function is particularly useful in deep learning
    workflows where resuming training from a specific epoch requires the RNG state
    to be consistent with that epoch's initial state (i.e. resuming dataloaders,
    batch norm, dropout states, etc.).

    Parameters
    ----------
    save_dir : str
        The directory where the RNG state file will be saved. The function will
        create a file named `rng_state_epoch_{epoch_index}.pt` in this directory.
    epoch_index : int
        The epoch index corresponding to the RNG state being saved. This index is
        used to name the saved file, indicating the epoch with which the RNG state
        is associated.
    """
    rng_state = {
        "python_random_state": random.getstate(),
        "numpy_random_state": np.random.get_state(),  # noqa: NPY002
        "torch_random_state": torch.random.get_rng_state(),
        "epoch_index": epoch_index,
    }
    if torch.cuda.is_available() and torch.cuda.is_initialized():  # type: ignore[no-untyped-call]
        # This will not be compatible with model parallelism
        rng_state["cuda"] = torch.cuda.get_rng_state()

    torch.save(rng_state, f"{save_dir}/rng_state_epoch_{epoch_index}.pt")


def load_and_set_rng_state(rng_state_path: str) -> Dict[str, Any]:
    """Load and set the state of various random number generators (RNGs) from a file,
    allowing the continuation of experiments or training processes from a specific
    point with reproducibility. This function restores the RNG states for Python's
    built-in `random` module, NumPy's random module, and PyTorch's global RNG,
    including CUDA's RNG state if applicable and present in the saved state.

    Parameters
    ----------
    rng_state_path : str
        The path to the file containing the saved RNG states. This file is expected
        to have been created by the `save_rng_state` function.

    Returns
    -------
    rng_state: Dict[str, Any]
        A dictionary containing the loaded RNG states. This includes states for
        Python's built-in `random` module (`python_random_state`), NumPy's random
        module (`numpy_random_state`), PyTorch's global RNG (`torch_random_state`),
        and, if applicable, PyTorch's CUDA RNG state (`cuda_rng_state`).
    """
    rng_state: Dict[str, Any] = torch.load(rng_state_path)

    random.setstate(rng_state["python_random_state"])
    np.random.set_state(rng_state["numpy_random_state"])  # noqa: NPY002
    torch.random.set_rng_state(rng_state["torch_random_state"])
    if "cuda_rng_state" in rng_state and torch.cuda.is_available():
        torch.cuda.set_rng_state(rng_state["cuda_rng_state"])
    return rng_state



"""This module contains utilities for seeding the random number generators to
ensure reproducibility.

Notes
-----
Reproducibility in deep learning ensures that experiments can be repeated with
identical results, critical for verifying research findings and deploying
reliable models. Distributed training introduces complexity because it involves
multiple computation units which may not synchronize their random states
perfectly. If training is paused and resumed, ensuring each unit starts with the
correct seed to reproduce the exact computational path becomes challenging. To
address this, one can find more sophisticated examples in libraries like
Composer, where the whole library's core is built around training deep neural
nets in any environment (distributed or not) with reproducibility in mind.

References
----------

- `PyTorch Reproducibility <https://pytorch.org/docs/stable/notes/randomness.html>`_
- `PyTorch deterministic algorithms <https://pytorch.org/docs/stable/generated/torch.use_deterministic_algorithms.html>`_
- `CUBLAS reproducibility <https://docs.nvidia.com/cuda/cublas/index.html#cublasApi_reproducibility>`_
- `PyTorch Worker <https://pytorch.org/docs/stable/notes/randomness.html#dataloader>`_
- `Composer <https://github.com/mosaicml/composer/blob/dev/composer/utils/reproducibility.py>`_
"""

from __future__ import annotations

import os
import random
import warnings

import numpy as np
import torch
import torch.backends.cudnn

__all__ = [
    "seed_all",
    "seed_worker",
    "configure_deterministic_mode",
    "raise_error_if_seed_is_negative_or_outside_32_bit_unsigned_integer",
]

max_seed_value = np.iinfo(np.uint32).max
min_seed_value = np.iinfo(np.uint32).min


def raise_error_if_seed_is_negative_or_outside_32_bit_unsigned_integer(value: int) -> None:
    if not (min_seed_value <= value <= max_seed_value):
        raise ValueError(f"Seed must be within the range [{min_seed_value}, {max_seed_value}]")


def configure_deterministic_mode() -> None:
    r"""
    Activates deterministic mode in PyTorch and CUDA to ensure reproducible
    results at the cost of performance and potentially higher CUDA memory usage.
    It sets deterministic algorithms, disables cudnn benchmarking and enables,
    and sets the CUBLAS workspace configuration.

    References
    ----------
    - `PyTorch Reproducibility <https://pytorch.org/docs/stable/notes/randomness.html>`_
    - `PyTorch deterministic algorithms <https://pytorch.org/docs/stable/generated/torch.use_deterministic_algorithms.html>`_
    - `CUBLAS reproducibility <https://docs.nvidia.com/cuda/cublas/index.html#cublasApi_reproducibility>`_
    """

    # fmt: off
    torch.use_deterministic_algorithms(True, warn_only=True)
    torch.backends.cudnn.benchmark        = False
    torch.backends.cudnn.deterministic    = True
    torch.backends.cudnn.enabled          = False

    os.environ['CUBLAS_WORKSPACE_CONFIG'] = ':4096:8'
    # fmt: on
    warnings.warn(
        "Deterministic mode is activated. This will negatively impact performance and may cause increase in CUDA memory footprint.",
        category=UserWarning,
        stacklevel=2,
    )


def seed_all(
    seed: int = 1992,
    seed_torch: bool = True,
    set_torch_deterministic: bool = True,
) -> int:
    """
    Seeds all relevant random number generators to ensure reproducible
    outcomes. Optionally seeds PyTorch and activates deterministic
    behavior in PyTorch based on the flags provided.

    Parameters
    ----------
    seed : int, default=1992
        The seed number for reproducibility.
    seed_torch : bool, default=True
        If True, seeds PyTorch's RNGs.
    set_torch_deterministic : bool, default=True
        If True, activates deterministic mode in PyTorch.

    Returns
    -------
    seed : int
        The seed number used for reproducibility.
    """
    raise_error_if_seed_is_negative_or_outside_32_bit_unsigned_integer(seed)

    # fmt: off
    os.environ["PYTHONHASHSEED"] = str(seed)       # set PYTHONHASHSEED env var at fixed value
    np.random.seed(seed)                           # numpy pseudo-random generator
    random.seed(seed)                              # python's built-in pseudo-random generator

    if seed_torch:
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)           # pytorch (both CPU and CUDA)
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.enabled = False

        if set_torch_deterministic:
            configure_deterministic_mode()
    # fmt: on
    return seed


def seed_worker(worker_id: int) -> None:  # noqa: ARG001
    """
    Initializes random seeds for a worker based on its ID.

    This ensures that each worker, when used in parallel data loading,
    operates with a unique random state, which enhances reproducibility
    and reducing potential data overlap or bias in data loading processes.

    Parameters
    ----------
    worker_id : int
        The unique identifier for the data loader worker.

    References
    ----------
    - `PyTorch Worker <https://pytorch.org/docs/stable/notes/randomness.html#dataloader>`_

    Examples
    --------
    .. code-block:: python

        import torch
        from torch.utils.data import DataLoader

        train_dataset = ...

        g = torch.Generator()
        g.manual_seed(0)

        DataLoader(
            train_dataset,
            batch_size=batch_size,
            num_workers=num_workers,
            worker_init_fn=seed_worker,
            generator=g,
        )
    """
    worker_seed = torch.initial_seed() % 2**32
    np.random.seed(worker_seed)  # noqa: NPY002
    random.seed(worker_seed)



from __future__ import annotations

import math
from typing import Literal

import torch
from torch import nn

from omnivault.modules.module import Activation


class Softmax(Activation):
    r"""Softmax activation function.

    Notes
    -----
    Let :math:`z` be of shape :math:`(B, K)` where :math:`B` is the batch size
    and :math:`K` is the number of classes. Then if we set :math:`\text{self.dim} = 1`,
    this means that the softmax function will be applied along the second dimension
    of :math:`z`. The output :math:`g` will be of shape :math:`(B, K)` and each row
    will sum to 1.

    In reality, we can have more than 2 dimensions (i.e., softmax in transformer
    model). Furthermore, softmax can be applied on feature dimensions as well
    , using a class is just convenient for this example.

    Stable Version
    --------------
    .. code-block:: python

        def stable_softmax(z: torch.Tensor) -> torch.Tensor:
            max_z = torch.max(z, dim=self.dim, keepdim=True).values
            numerator = torch.exp(z - max_z)
            denominator = torch.sum(numerator, dim=self.dim, keepdim=True)
            g = numerator / denominator
            return g

    Examples
    --------
    We can easily see that given a tensor :math:`\mathbf{Z}` of shape :math:`(B, K)= (2, 3)`,
    and if we specify :math:`\text{dim} = -1`, the softmax function will be applied
    along the last dimension of :math:`\mathbf{Z}`. In this case, the softmax will be applied
    to each row of :math:`\mathbf{Z}` and the output :math:`g` will be of shape :math:`(2, 3)`.
    We can easily verify that softmax is indeed applied independently to each row of :math:`\mathbf{Z}`
    by comparing the output of the softmax function applied to each row of :math:`\mathbf{Z}`
    with the output of the softmax function applied to the entire tensor :math:`\mathbf{Z}`.

    >>> softmax = Softmax(dim=-1)
    >>> Z = torch.tensor([[1.0, 2.0, 3.0], [3.0, 2.0, 1.0]])
    >>> result = softmax(Z)
    >>> result
    tensor([[0.0900, 0.2447, 0.6652],
            [0.6652, 0.2447, 0.0900]])
    >>> individual_results = torch.vstack([softmax(z[i].unsqueeze(0)) for i in range(z.shape[0])])
    >>> torch.testing.assert_close(result, individual_results)
    """

    def __call__(self, z: torch.Tensor) -> torch.Tensor:
        r"""
        Compute the softmax function for a given input.

        Parameters
        ----------
        z : torch.Tensor
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output of the softmax function.
        """
        numerator = torch.exp(z)
        denominator = torch.sum(numerator, dim=self.dim, keepdim=True)
        g = numerator / denominator
        return g

    def gradient(self, z: torch.Tensor) -> torch.Tensor:
        r"""
        Use the Jacobian matrix to compute the gradient.

        For usage, see:
        :file:`omniverse/playbook/softmax_preserves_order_translation_invariant_not_invariant_scaling.md`.

        Parameters
        ----------
        z : torch.Tensor
            Input tensor.

        Returns
        -------
        torch.Tensor
            Gradient of the softmax function.
        """
        S = self.__call__(z)
        diag_S = torch.diag_embed(S)
        outer_S = torch.matmul(S.unsqueeze(2), S.unsqueeze(1))
        gradient = diag_S - outer_S
        return gradient.squeeze()


class SoftmaxStable(Softmax):
    def __call__(self, z: torch.Tensor) -> torch.Tensor:
        max_z = torch.max(z, dim=self.dim, keepdim=True).values
        numerator = torch.exp(z - max_z)
        denominator = torch.sum(numerator, dim=self.dim, keepdim=True)
        g = numerator / denominator
        return g


class GELU(nn.Module):
    r"""A Gaussian Error Linear Unit (GELU) activation layer that applies the GELU function element-wise.

    The GELU activation function is defined as:

    .. math::
        \text{GELU}(x) = x \cdot \Phi(x)

    where :math:`\Phi(x)` is the cumulative distribution function of the standard Gaussian distribution.

    Parameters
    ----------
    approximate : {'tanh', None}, optional
        Specifies whether to use an approximation of the GELU function:
        - 'tanh': Uses a tanh-based approximation to compute GELU.
        - None: Uses the exact formula with the error function `erf`.
        The default is None, which uses the exact formulation.

    Notes
    -----
    The GELU function is mathematically represented as:

    .. math::
        \text{GELU}(x) = x \cdot \frac{1}{2} \left[1 + \text{erf}\left(\frac{x}{\sqrt{2}}\right)\right]

    When GELU (Gaussian Error Linear Unit) is applied to a tensor, it operates
    element-wise, meaning it is applied independently and identically (i.i.d.) to
    each element within the tensor. Given a tensor with shape
    :math:`[\mathcal{B}, T, D] = [2, 3, 4]`, which represents a batch size of 2,
    sequence length of 3, and feature dimension of 4, the GELU activation function
    would be applied 24 times in total, once for each element in the tensor.

    The `approximate='tanh'` option applies the following approximation:

    .. math::
        \text{GELU}_{\text{approx}}(x) = 0.5x \cdot \left(1 + \text{tanh}\left(\sqrt{\frac{2}{\pi}} \cdot \left(x + 0.044715 \cdot x^3 \right)\right)\right)

    This approximation is computationally faster and has been found to perform similarly to the exact form.

    Examples
    --------
    The GELU activation function is applied to a tensor :math:`\mathbf{Z}` of shape :math:`(2, 3)`
    elementwise, we can easily verify with an assert statement below to show that applying
    GELU on a tensor is equivalent to applying the GELU function elementwise to each element
    in the tensor.

    >>> import torch
    >>> Z = torch.tensor([[-1, 0, 1], [1, 0, -1]], dtype=torch.float32)
    >>> gelu = GELU(approximate=None)
    >>> gelu(Z)
    tensor([[-0.1587,  0.0000,  0.8413],
            [ 0.8413,  0.0000, -0.1587]])
    >>> for i, row in enumerate(Z):
    ...     for j, element in enumerate(row):
    ...         out = GELU()(element)
    ...         torch.testing.assert_close(out, result[i, j])
    """

    def __init__(self, approximate: Literal["tanh"] | None = None) -> None:
        super().__init__()
        self.approximate = approximate

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.approximate == "tanh":
            x_out_BTD = 0.5 * x * (1.0 + torch.tanh(math.sqrt(2.0 / math.pi) * (x + 0.044715 * torch.pow(x, 3.0))))
            return x_out_BTD

        x_out_BTD = x * 0.5 * (1 + torch.erf(input=x / 2**0.5))
        return x_out_BTD



"""Whether to pull the `generate` method out from the `DecoderGPT` class
and make it a standalone function/class is a design issue. Putting it under
the `DecoderGPT` class makes it easier to access the `self` attributes."""
